/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <limits.h>
#include <cstring>
#include <list>
#include <stack>
#include <map>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <sstream>
#include <assert.h>
#include "lexgen.h"

#include "lexskel.inc"

using namespace std;

LexGen::Rule::Rule(int id, const NFA& nfa, const string& action, int scond, const NFA* trailer, int options)
:	m_id(id), m_action(action), m_scond(scond), m_options(options)
{
	m_nfa = nfa.clone();
	m_trailer = trailer ? trailer->clone() : 0;
	m_main_length = nfa.length();
	m_trailer_length = trailer ? trailer->length() : 0;
}

LexGen::Rule::Rule(const Rule& rule)
{
	m_nfa = rule.m_nfa->clone();
	m_id = rule.m_id;
	m_action = rule.m_action;
	m_scond = rule.m_scond;
	m_trailer = rule.m_trailer ? rule.m_trailer->clone() : 0;
	m_options = rule.m_options;
	m_main_length = rule.m_main_length;
	m_trailer_length = rule.m_trailer_length;
}

LexGen::Rule::~Rule()
{
	delete m_nfa;
	if (m_trailer)
		delete m_trailer;
}

LexGen::Rule&
LexGen::Rule::operator=(const Rule& rule)
{
	delete m_nfa;
	if (m_trailer)
		delete m_trailer;
	m_nfa = rule.m_nfa->clone();
	m_id = rule.m_id;
	m_action = rule.m_action;
	m_scond = rule.m_scond;
	m_trailer = rule.m_trailer ? rule.m_trailer->clone() : 0;
	m_options = rule.m_options;
	m_main_length = rule.m_main_length;
	m_trailer_length = rule.m_trailer_length;
	return *this;
}

void
LexGen::addRule(int scond, const NFA& nfa, const string& action, const NFA* trailer, int options)
{
	Rule rule(m_rule_id_generator++, nfa, action, scond, trailer, options);
	m_rules_by_id.insert(rules_by_id_type::value_type(rule.m_id, rule));
	rules_by_scond_type::iterator fiter = m_rules_by_scond.find(scond);
	if (fiter != m_rules_by_scond.end()) {
		fiter->second.push_back(rule);
	} else {
		m_rules_by_scond.insert(rules_by_scond_type::value_type(scond, list<Rule>(1, rule)));
	}
}

void
LexGen::generateLexer(ostream& out)
{
	if (m_rules_by_scond.empty()) {
		cerr << "lexgen: nothing to do" << endl;
		throw 1;
	}
	rules_by_scond_type::const_iterator iter = m_rules_by_scond.begin();
	int min_scond = INT_MAX;
	int max_scond = INT_MIN;
	for (; iter != m_rules_by_scond.end(); ++iter) {
		const int scond = iter->first;
		if (scond <= min_scond)
			min_scond = scond;
		if (scond >= max_scond)
			max_scond = scond;
	}
	if (min_scond != 0) {
		cerr << "lexgen: Start conditions must start with zero" << endl;
		throw 1;
	}
	for (int i = min_scond; i <= max_scond; i++) {
		if ((iter = m_rules_by_scond.find(i)) == m_rules_by_scond.end()) {
			cerr << "lexgen: Start condtions must not have gaps" << endl;
			throw 1;
		}
	}
	
	int n_rules = m_rules_by_id.size();
	
	ostringstream actions_fn;
	if (!m_use_subclass.empty()) {
		actions_fn << "int\nLexer::executeAction(short rule)\n{\n\treturn 0;\n}\n\n";
		actions_fn << "int\n" << m_use_subclass << "::executeAction(short rule)\n{\n";
	} else {
		actions_fn << "int\nLexer::executeAction(short rule)\n{\n";
	}
	vector<string> rule_code(n_rules);
	int middle = buildGotoTree(rule_code, 0, n_rules);
	actions_fn << "\twhile(true) {\n\t\tgoto rule" << middle << ";\n";
	for (int i = 0; i < n_rules; i++) {
		actions_fn << rule_code[i];
	}
	actions_fn << "\t}\n\treturn 0;\n}\n";
	
	ostringstream dtran_tbl, dtran_tbls, total_states_tbl, non_deadend_states_tbl;
	ostringstream stateprop_tbl, stateprop_tbls, stateinfo_tbl;
	unsigned int stateinfo_off = 0;
	dtran_tbl << "const short *Lexer::m_dtran[] = {" << endl;
	total_states_tbl << "const short Lexer::m_total_states[] = {" << endl;
	non_deadend_states_tbl << "const short Lexer::m_non_deadend_states[] = {" << endl;
	stateprop_tbl << "const unsigned int *Lexer::m_stateprop[] = {" << endl;
	stateinfo_tbl << "const short Lexer::m_stateinfo[] = {";
	iter = m_rules_by_scond.begin();
	for (; iter != m_rules_by_scond.end(); ++iter) {
		const int scond = iter->first;
		LexerRuleComposition composition(iter->second);
		
		if (scond != 0) {
			dtran_tbl << ", ";
			stateprop_tbl << ", ";
			total_states_tbl << ", ";
			non_deadend_states_tbl << ", ";
			if (scond % 8 == 0) {
				dtran_tbl << endl;
				stateprop_tbl << endl;
			}
			if (scond % 16 == 0) {
				total_states_tbl << endl;
				non_deadend_states_tbl << endl;
			}
		}
		
		dtran_tbls << "static const short dtran" << scond << "[] = {";
		stateprop_tbls << "static const unsigned int stateprop" << scond << "[] = {";
		
		dtran_tbl << "dtran" << scond;
		stateprop_tbl << "stateprop" << scond;
		
		list<DState> dstates;
		vector<DState> processed_dstates;
		dstates.push_back(eClosureInitial(composition));
		dstates.rbegin()->m_state_num = 0;
		int counter = 1;
		while (!dstates.empty()) {
			DState dstate = *dstates.begin();
			dstates.pop_front();
			int finish_rule = INT_MAX;
			bool is_quickfinish = false;
			bool is_lazy_finish = false;
			typedef map<int, DStateInfo> ds_info_type;
			ds_info_type ds_info;
			set<int>::const_iterator iter2 = dstate.m_stateset.begin();
			for (; iter2 != dstate.m_stateset.end(); ++iter2) {
				if (*iter2 < 0)
					continue;
				
				NStateInfo ns_info = composition.getStateInfo(*iter2);
				
				if (ns_info.m_is_trailer) {
					ds_info[ns_info.m_rule].m_nTrailerStates++;
				} else {
					ds_info[ns_info.m_rule].m_nMainStates++;
				}
				
				if (ns_info.m_is_semifinish) {
					dstate.m_semifinish_for.insert(ns_info.m_rule);
				}
				
				if (ns_info.m_is_quickfinish || ns_info.m_is_longfinish) {
					if (ns_info.m_rule < finish_rule) {
						finish_rule = ns_info.m_rule;
						is_quickfinish = ns_info.m_is_quickfinish;
						is_lazy_finish = ns_info.m_is_lazy;
					}
				}
			}
			if (finish_rule < INT_MAX) {
				dstate.m_finish_for = finish_rule;
				dstate.m_is_quickfinish = is_quickfinish;
				dstate.m_is_lazy_finish = is_lazy_finish;
			}
			ds_info_type::const_iterator ds_iter = ds_info.begin();
			for (; ds_iter != ds_info.end(); ++ds_iter) {
				if (ds_iter->second.m_nMainStates > 0 && ds_iter->second.m_nTrailerStates > 0) {
					cerr << "lexgen: Dangerous trailing context for rule #" << ds_iter->first << ":" << endl
						<< "the beginning of the trailing context matches the end of the main pattern." << endl;
					rules_by_id_type::const_iterator fit = m_rules_by_id.find(ds_iter->first);
					assert(fit != m_rules_by_id.end());
					int main_length = fit->second.m_main_length;
					int trailer_length = fit->second.m_trailer_length;
					if (main_length == -1 && trailer_length == -1) {
						cerr << "Both main pattern and the trailing context have variable length - "
							<< endl << "no way to handle it." << endl;
					} else {
						if (main_length != -1) {
							cerr << "The main pattern has fixed length";
						} else {
							cerr << "The trailing context has fixed length";
						}
						cerr << " - possible to handle, " << "but not implemented." << endl
							<< "Consider a workaround using less()" << endl;
					}
					throw 1;
				}
			}
			bool dead_end = true;
			for (int ch = 0; ch < 256; ch++) {
				DState new_dstate = eClosure(move(dstate, ch));
				if (new_dstate.m_stateset.empty()) {
					dstate.m_moves[ch] = -1;
					continue;
				}
				const DState *dst_dstate;
				list<DState>::const_iterator fiter1 = find(dstates.begin(), dstates.end(), new_dstate);
				vector<DState>::const_iterator fiter2 = find(processed_dstates.begin(), processed_dstates.end(), new_dstate);
				if (fiter1 != dstates.end()) {
					dst_dstate = &(*fiter1);
				} else if (fiter2 != processed_dstates.end()) {
					dst_dstate = &(*fiter2);
				} else if (new_dstate == dstate) {
					dst_dstate = &dstate;
				} else {
					new_dstate.m_state_num = counter++;
					dstates.push_back(new_dstate);
					dst_dstate = &new_dstate;
				}
				dstate.m_moves[ch] = dst_dstate->m_state_num;
				dead_end = false;
			}
			dstate.m_dead_end = dead_end;
			//cerr << "DState #" << dstate.m_state_num << ": is_deadend=" << dead_end << endl;
			processed_dstates.push_back(dstate);
		}
		
		map<int,int,less<int> > dstate2out_map;
		map<int,int,less<int> > out2dstate_map;
		vector<DState>::const_iterator iter3 = processed_dstates.begin();
		int cnt = 0;
		for (; iter3 != processed_dstates.end(); ++iter3) {
			if (iter3->m_dead_end)
				continue;
			//cerr << "DState " << cnt << " was #" << iter3->m_state_num << endl;
			dstate2out_map[iter3->m_state_num] = cnt;
			out2dstate_map[cnt] = iter3->m_state_num;
			cnt++;
		}
		int non_deadend_states = cnt;
		iter3 = processed_dstates.begin();
		for (; iter3 != processed_dstates.end(); ++iter3) {
			if (!iter3->m_dead_end)
				continue;
			//cerr << "DState " << cnt << " was #" << iter3->m_state_num << endl;
			dstate2out_map[iter3->m_state_num] = cnt;
			out2dstate_map[cnt] = iter3->m_state_num;
			cnt++;
		}
		int total_states = cnt;
		
		for (int i = 0; i < total_states; i++) {
			/*
			stateprop layout
			bit 0: semifinish
			bit 1: quick finish
			bit 2: lazy finish
			bit 3: finish
			bits 4-32: offset in stateinfo table
			*/
			const DState& dstate = processed_dstates[out2dstate_map[i]];
			unsigned int stateprop = 0;
			unsigned int off = stateinfo_off;
			if (dstate.m_finish_for >= 0) {
				if (stateinfo_off > 0)
					stateinfo_tbl << ", ";
				if (stateinfo_off % 16 == 0)
					stateinfo_tbl << endl;
				stateinfo_tbl << dstate.m_finish_for;
				stateinfo_off++;
				stateprop |= 8;
				stateprop |= off << 4;
				if (dstate.m_is_quickfinish) {
					stateprop |= 2;
				}
				if (dstate.m_is_lazy_finish) {
					stateprop |= 4;
				}
			}
			if (!dstate.m_semifinish_for.empty()) {
				set<int>::const_iterator iter = dstate.m_semifinish_for.begin();
				for (; iter != dstate.m_semifinish_for.end(); ++iter) {
					if (stateinfo_off > 0)
						stateinfo_tbl << ", ";
					if (stateinfo_off % 16 == 0)
						stateinfo_tbl << endl;
					stateinfo_tbl << *iter;
					stateinfo_off++;
				}
				stateinfo_tbl << ", -1";
				stateinfo_off++;
				stateprop |= 1;
				stateprop |= off << 4;
			}
			if (i != 0) {
				stateprop_tbls << ", ";
			}
			if (i % 8 == 0)
				stateprop_tbls << endl;
			stateprop_tbls << stateprop;
		}
		stateprop_tbls << endl << "};" << endl;
		
		bool first = true;
		cnt = 0;
		for (int i = 0; i < non_deadend_states; i++) {
			short *moves = processed_dstates[out2dstate_map[i]].m_moves;
			for (int ch = 0; ch < 256; ch++) {
				if (!first) {
					dtran_tbls << ", ";
				} else {
					first = false;
				}
				if (cnt++ % 16 == 0)
					dtran_tbls << endl;
				if (moves[ch] < 0) {
					dtran_tbls << moves[ch];
				} else {
					dtran_tbls << dstate2out_map[moves[ch]];
				}
			}
		}
		dtran_tbls << endl << "};" << endl;
		total_states_tbl << total_states;
		non_deadend_states_tbl << non_deadend_states;
	}
	dtran_tbl << endl << "};" << endl;
	total_states_tbl << endl << "};" << endl;
	non_deadend_states_tbl << endl << "};" << endl;
	stateprop_tbl << endl << "};" << endl;
	stateinfo_tbl << endl << "};" << endl;
	if (!m_include_code.empty()) {
		out << m_include_code << endl;
	}
	for (int i = 0; skel[i]; i++) {
		out << skel[i] << endl;
	}
	out << "void\nLexer::initSemifinishBuffer()\n{\n";
	out << "\tm_semifinishes = new int[" << n_rules << "];\n";
	out << "\tmemset(m_semifinishes, 0, sizeof(int)*" << n_rules << ");\n}\n\n";
	out << "void\nLexer::destroySemifinishBuffer()\n{\n";
	out << "\tdelete[] m_semifinishes;\n}\n\n";
	out << actions_fn.str();
	out << total_states_tbl.str();
	out << non_deadend_states_tbl.str();
	out << dtran_tbls.str();
	out << dtran_tbl.str();
	out << stateprop_tbls.str();
	out << stateprop_tbl.str();
	out << stateinfo_tbl.str();
	return;
}

int
LexGen::buildGotoTree(vector<string>& rule_code, int from, int to)
{
	if (to == from) {
		return from;
	} else if (to == from+1) {
		rules_by_id_type::const_iterator fit = m_rules_by_id.find(from);
		assert(fit != m_rules_by_id.end());
		ostringstream strm;
		strm << "\trule" << from << ":\n\t\t" << fit->second.m_action << "\n\t\tbreak;\n";
		rule_code[from] = strm.str();
		return from;
	} else if (to == from+2) {
		buildGotoTree(rule_code, from, from+1);
		rules_by_id_type::const_iterator fit = m_rules_by_id.find(from+1);
		assert(fit != m_rules_by_id.end());
		ostringstream strm;
		strm << "\trule" << from+1 << ":\n";
		strm << "\t\tif (rule < " << from+1 << ") goto rule" << from << ";\n";
		strm << "\t\t" << fit->second.m_action << "\n\t\tbreak;\n";
		rule_code[from+1] = strm.str();
		return from+1;
	} else {
		int middle = from+(to-from)/2;
		int middle1 = buildGotoTree(rule_code, from, middle);
		int middle2 = buildGotoTree(rule_code, middle+1, to);
		rules_by_id_type::const_iterator fit = m_rules_by_id.find(middle);
		assert(fit != m_rules_by_id.end());
		ostringstream strm;
		strm << "\trule" << middle << ":\n";
		strm << "\t\tif (rule < " << middle << ") goto rule" << middle1 << ";\n";
		strm << "\t\tif (rule > " << middle << ") goto rule" << middle2 << ";\n";
		strm << "\t\t" << fit->second.m_action << "\n\t\tbreak;\n";
		rule_code[middle] = strm.str();
		return middle;
	}
	
}

LexGen::DState
LexGen::eClosureInitial(const LexerRuleComposition& composition)
{
	DState res(&composition);
	stack<int> nfa_states;
	nfa_states.push(NFA::START_STATE);
	res.m_stateset.insert(NFA::START_STATE);
	while (!nfa_states.empty()) {
		int nstate = nfa_states.top();
		nfa_states.pop();
		list<NFAMove> moves = composition.getMovesFrom(nstate);
		list<NFAMove>::const_iterator iter = moves.begin();
		for (; iter != moves.end(); ++iter) {
			if (!iter->m_is_emove)
				continue;
			if (res.m_stateset.find(iter->m_move_to) == res.m_stateset.end()) {
				res.m_stateset.insert(iter->m_move_to);
				nfa_states.push(iter->m_move_to);
			}
		}
	}
	return res;
}

LexGen::DState
LexGen::eClosure(const DState& dstate)
{
	DState res(dstate.m_composition);
	stack<int> nfa_states;
	set<int>::const_iterator it = dstate.m_stateset.begin();
	for (; it != dstate.m_stateset.end(); ++it) {
		nfa_states.push(*it);
		res.m_stateset.insert(*it);
	}
	while (!nfa_states.empty()) {
		int nstate = nfa_states.top();
		nfa_states.pop();
		list<NFAMove> moves = dstate.m_composition->getMovesFrom(nstate);
		list<NFAMove>::const_iterator iter = moves.begin();
		for (; iter != moves.end(); ++iter) {
			if (!iter->m_is_emove)
				continue;
			if (res.m_stateset.find(iter->m_move_to) == res.m_stateset.end()) {
				res.m_stateset.insert(iter->m_move_to);
				nfa_states.push(iter->m_move_to);
			}
		}
	}
	return res;
}

LexGen::DState
LexGen::move(DState& dstate, char ch)
{
	DState res(dstate.m_composition);
	set<int>::const_iterator it = dstate.m_stateset.begin();
	for (; it != dstate.m_stateset.end(); ++it) {
		list<NFAMove> moves = dstate.m_composition->getMovesFrom(*it);
		list<NFAMove>::const_iterator iter = moves.begin();
		for (; iter != moves.end(); ++iter) {
			if (iter->m_is_emove)
				continue;
			if (iter->m_char == ch) {
				res.m_stateset.insert(iter->m_move_to);
			}
		}
	}
	return res;
}

LexGen::LexerRuleComposition::LexerRuleComposition(const list<Rule>& rules)
:	m_nstates(0)
{
	list<Rule>::const_iterator iter = rules.begin();
	for (; iter != rules.end(); iter++) {
		int rule_id = iter->m_id;
		bool rule_lazy = (iter->m_options & OPT_LAZY);
		const NFA *nfa = iter->m_nfa;
		int new_start = newState(rule_id, rule_lazy);
		int new_finish = newState(rule_id, rule_lazy);
		list<NFAMove> moves = nfa->getMovesFrom(NFA::START_STATE);
		list<NFAMove>::iterator iter2 = moves.begin();
		for (; iter2 != moves.end(); ++iter2) {
			if (iter2->m_move_to == NFA::FINISH_STATE) {
				iter2->m_move_to = new_finish;
			} else {
				iter2->m_move_to += m_nstates;
			}
		}
		m_moves.push_back(moves);
		m_moves.push_back(list<NFAMove>());
		m_start_moves.push_back(new_start);
		int states_offset = m_nstates;
		int nstates = nfa->nStates();
		for (int i = 0; i < nstates; i++) {
			newState(rule_id, rule_lazy);
			list<NFAMove> moves = nfa->getMovesFrom(i);
			iter2 = moves.begin();
			for (; iter2 != moves.end(); ++iter2) {
				if (iter2->m_move_to == NFA::FINISH_STATE) {
					iter2->m_move_to = new_finish;
				} else {
					iter2->m_move_to += states_offset;
				}
			}
			m_moves.push_back(moves);
		}
		
		if (!iter->m_trailer) {
			stateinfo_type::iterator fit = m_stateinfo.find(new_finish);
			assert(fit != m_stateinfo.end());
			fit->second.m_is_quickfinish = true;
			continue;
		}
		
		const NFA *trailer = iter->m_trailer;
		int semifinish = new_finish;
		new_finish = newState(rule_id, rule_lazy);
		m_moves.push_back(list<NFAMove>());
		list<NFAMove> moves2 = trailer->getMovesFrom(NFA::START_STATE);
		iter2 = moves2.begin();
		for (; iter2 != moves2.end(); ++iter2) {
			if (iter2->m_move_to == NFA::FINISH_STATE) {
				iter2->m_move_to = new_finish;
			} else {
				iter2->m_move_to += m_nstates;
			}
		}
		m_moves[semifinish] = moves2;
		states_offset = m_nstates;
		nstates = trailer->nStates();
		for (int i = 0; i < nstates; i++) {
			newState(rule_id, rule_lazy);
			list<NFAMove> moves = trailer->getMovesFrom(i);
			iter2 = moves.begin();
			for (; iter2 != moves.end(); ++iter2) {
				if (iter2->m_move_to == NFA::FINISH_STATE) {
					iter2->m_move_to = new_finish;
				} else {
					iter2->m_move_to += states_offset;
				}
			}
			m_moves.push_back(moves);
		}
		
		stateinfo_type::iterator sf_it = m_stateinfo.find(semifinish);
		assert(sf_it != m_stateinfo.end());
		stateinfo_type::iterator nf_it = m_stateinfo.find(new_finish);
		assert(nf_it != m_stateinfo.end());
		sf_it->second.m_is_semifinish = true;
		nf_it->second.m_is_longfinish = true;
		nf_it->second.m_is_trailer = true;
	}
}

list<NFAMove>
LexGen::LexerRuleComposition::getMovesFrom(int state) const
{
	if (state == NFA::FINISH_STATE) {
		return list<NFAMove>();
	} else if (state == NFA::START_STATE) {
		return m_start_moves;
	} else {
		return m_moves[state];
	}
}

LexGen::NStateInfo
LexGen::LexerRuleComposition::getStateInfo(int state) const
{
	stateinfo_type::const_iterator fit = m_stateinfo.find(state);
	assert(fit != m_stateinfo.end());
	return fit->second;
}

int
LexGen::LexerRuleComposition::newState(int rule, bool lazy)
{
	m_stateinfo.insert(stateinfo_type::value_type(m_nstates, NStateInfo(rule, lazy)));
	return m_nstates++;
}

LexGen::DState::DState(const LexerRuleComposition* composition)
:	m_composition(composition), m_state_num(-1), m_finish_for(-1),
	m_is_quickfinish(false), m_is_lazy_finish(false), m_dead_end(false),
	m_moves(new short[256])
{}

LexGen::DState::DState(const DState& dstate)
:	m_composition(dstate.m_composition), m_moves(new short[256])
{
	m_stateset = dstate.m_stateset;
	m_state_num = dstate.m_state_num;
	m_finish_for = dstate.m_finish_for;
	m_is_quickfinish = dstate.m_is_quickfinish;
	m_is_lazy_finish = dstate.m_is_lazy_finish;
	m_semifinish_for = dstate.m_semifinish_for;
	m_dead_end = dstate.m_dead_end;
	memcpy(m_moves, dstate.m_moves, sizeof(short[256]));
}

LexGen::DState&
LexGen::DState::operator=(const DState& dstate)
{
	m_composition = dstate.m_composition;
	m_stateset = dstate.m_stateset;
	m_state_num = dstate.m_state_num;
	m_finish_for = dstate.m_finish_for;
	m_is_quickfinish = dstate.m_is_quickfinish;
	m_is_lazy_finish = dstate.m_is_lazy_finish;
	m_semifinish_for = dstate.m_semifinish_for;
	m_dead_end = dstate.m_dead_end;
	memcpy(m_moves, dstate.m_moves, sizeof(short[256]));
	return *this;
}

