/*
  regexec.c - TRE regex matching engine

  Copyright (C) 2001, 2002 Ville Laurikari <vl@iki.fi>.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 (June
  1991) as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

/* AIX requires this to be the first thing in the file.  */
#ifndef __GNUC__
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  ifdef _AIX
 #pragma alloca
#  else
#   ifndef alloca /* predefined by HP cc +Olibcalls */
char *alloca ();
#   endif
#  endif
# endif
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_WCHAR_H
#include <wchar.h>
#endif /* HAVE_WCHAR_H */
#ifdef HAVE_WCTYPE_H
#include <wctype.h>
#endif /* HAVE_WCTYPE_H */
#ifndef TRE_WCHAR
#include <ctype.h>
#endif /* !TRE_WCHAR */

#include "tre.h"
#include "regex.h"
#include "xmalloc.h"


typedef struct {
  tre_tnfa_transition_t *state;
  int *tags;
} tre_tnfa_reach_t;

typedef struct {
  int pos;
  int **tags;
} tre_reach_pos_t;


#ifdef TRE_DEBUG
static void
print_reach(const tre_tnfa_t *tnfa, tre_tnfa_reach_t *reach)
{
  int i;

  while (reach->state != NULL)
    {
      DPRINT((" %p", (void *)reach->state));
      if (tnfa->num_tags > 0)
	{
	  DPRINT(("/"));
	  for (i = 0; i < tnfa->num_tags; i++)
	    {
	      DPRINT(("%d:%d", i, reach->tags[i]));
	      if (i < (tnfa->num_tags-1))
		DPRINT((","));
	    }
	}
      reach++;
    }
  DPRINT(("\n"));

}
#endif /* TRE_DEBUG */

/* Returns 1 if `t1' wins `t2', 0 otherwise. */
inline static int
tag_order(int num_tags, int *tag_directions, int *t1, int *t2)
{
  int i;
  /*
  DPRINT(("tag_order\n"));
  DPRINT(("t1[0] = %d, t2[0] = %d\n", *t1, *t2));
  */
  i = 0;
  while (i < num_tags)
    {
      if (tag_directions[i] == TAG_MINIMIZE)
	{
	  DPRINT(("< t1[%d] = %d, t2[%d] = %d\n", i, t1[i], i, t2[i]));
	  if (t1[i] > t2[i])
	    return 0;
	  if (t1[i] < t2[i])
	    return 1;
	}
      else
	{
	  DPRINT(("> t1[%d] = %d, t2[%d] = %d\n", i, t1[i], i, t2[i]));
	  if (t1[i] > t2[i])
	    return 1;
	  if (t1[i] < t2[i])
	    return 0;
	}
      i++;
    }
  /*  assert(0);*/
  return 0;
}

inline static int
neg_char_classes_match(tre_ctype_t *classes, tre_cint_t wc, int icase)
{
  DPRINT(("neg_char_classes_test: %p, %d, %d\n", classes, wc, icase));
  while (*classes != (tre_ctype_t)0)
    if ((!icase && tre_isctype(wc, *classes))
	|| (icase && (tre_isctype(tre_toupper(wc), *classes)
		      || tre_isctype(tre_tolower(wc), *classes))))
      return 1; /* Match. */
    else
      classes++;
  return 0; /* No match. */
}

/* XXX - Make sure that this does not refer past the end of very short
         strings (length 0 - 4 chars).
   XXX - Make sure assertions work when len < 0.
   XXX - Use a type for `len' which can hold the maximum value of a `size_t'.
   XXX - Go through the code with a fine comb to spot bugs and optimization
         possibilities.  Fix/implement/document them. */
static reg_errcode_t
tnfa_execute(const tre_tnfa_t *tnfa, const void *string, int len,
	     tre_str_type_t type, int *match_tags, int eflags,
	     int *match_end_ofs)
{
  tre_char_t prev_c = 0, next_c = 0;
  const char *str_byte = string;
  char *buf;
  int pos, num_tags, pos_add = 1;
  tre_tnfa_transition_t *trans_i;
  tre_tnfa_reach_t *reach, *reach_next, *reach_i, *reach_next_i;
  tre_reach_pos_t *reach_pos;
  int *tag_i;
  int i;
  int reg_notbol = eflags & REG_NOTBOL;
  int reg_noteol = eflags & REG_NOTEOL;
  int reg_newline = tnfa->cflags & REG_NEWLINE;
  int match_eo = -1;       /* end offset of match (-1 if no match found yet) */
  int *tmp_tags = NULL;
  int *tmp_iptr;
  int cflags = tnfa->cflags;
#ifdef TRE_WCHAR
  const wchar_t *str_wide = string;
#ifdef TRE_MULTIBYTE
  int pos_add_next = 0;
#ifdef HAVE_MBSTATE_T
  mbstate_t mbstate;
  memset(&mbstate, '\0', sizeof(mbstate));
#endif /* HAVE_MBSTATE_T */
#endif /* TRE_MULTIBYTE */
#endif /* TRE_WCHAR */

  if (eflags & REG_NOTAGS)
    num_tags = 0;
  else
    num_tags = tnfa->num_tags;

#if defined(TRE_WCHAR) && defined(TRE_MULTIBYTE)
#define GET_NEXT_WCHAR()                                                      \
  do {                                                                        \
    prev_c = next_c;                                                          \
    if (type == STR_BYTE)					      	      \
      next_c = (unsigned char)(*str_byte++);			       	      \
    else if (type == STR_WIDE)						      \
      next_c = *str_wide++;			       			      \
    else if (type == STR_MBS)						      \
      {									      \
        size_t w;	                                                      \
        if (str_byte == NULL)                                                 \
          w = L'\0';                                                          \
        else                                                                  \
          {                                                                   \
            w = tre_mbrtowc(&next_c, str_byte, 32, &mbstate);		      \
            if (w < 0)                                                        \
              {	                                                              \
                xfree(buf);						      \
   	        return REG_NOMATCH;                                           \
              }                                                               \
	    pos_add = pos_add_next;                                           \
	    pos_add_next = w;                                                 \
            str_byte += w;                                                    \
          }                                                                   \
     }                                                                        \
  } while(0)
#elif defined(TRE_WCHAR)
#define GET_NEXT_WCHAR()                                                      \
  do {                  						      \
    prev_c = next_c;							      \
    if (type == STR_BYTE)						      \
      next_c = (unsigned char)(*str_byte++);				      \
    else if (type == STR_WIDE)						      \
      next_c = *str_wide++;                                                   \
  } while(0)
#else
#define GET_NEXT_WCHAR()                                                      \
  do {                                                                        \
    prev_c = next_c;							      \
    next_c = (unsigned char)(*str_byte++);                                    \
  } while(0)
#endif /* !defined(TRE_WCHAR) */

#define IS_WORD_CHAR(c)  ((c) == L'_' || tre_isalnum(c))

#define CHECK_ASSERTIONS(assertions)                                          \
  (((assertions & ASSERT_AT_BOL)					      \
    && (pos > 0 || reg_notbol)						      \
    && (prev_c != L'\n' || !reg_newline))				      \
   || ((assertions & ASSERT_AT_EOL)					      \
       && (next_c != L'\0' || reg_noteol)				      \
       && (next_c != L'\n' || !reg_newline))				      \
   || ((assertions & ASSERT_AT_BOW)					      \
       && (pos > 0 && (IS_WORD_CHAR(prev_c) || !IS_WORD_CHAR(next_c))))	      \
   || ((assertions & ASSERT_AT_EOW)					      \
       && (!IS_WORD_CHAR(prev_c) || IS_WORD_CHAR(next_c)))		      \
   || ((assertions & ASSERT_AT_WB)					      \
       && (pos != 0 && next_c != L'\0'					      \
	   && IS_WORD_CHAR(prev_c) == IS_WORD_CHAR(next_c)))		      \
   || ((assertions & ASSERT_AT_WB_NEG)					      \
       && (pos == 0 || next_c == L'\0'                                        \
	   || IS_WORD_CHAR(prev_c) != IS_WORD_CHAR(next_c))))

  DPRINT(("tnfa_execute, input type %d\n", type));

  /* Allocate memory for temporary data required for matching.  This needs to
     be done for every matching operation to be thread safe.  This allocates
     everything in a single large block from the stack frame using alloca(). */
  {
    int tbytes, rbytes, pbytes, xbytes, total_bytes;
    char *tmp_buf;
    tbytes = sizeof(*tmp_tags) * num_tags;
    tbytes += tbytes % sizeof(long);
    rbytes = sizeof(*reach_next) * (tnfa->num_states + 1);
    rbytes += rbytes % sizeof(long);
    pbytes = sizeof(*reach_pos) * tnfa->num_states;
    pbytes += pbytes % sizeof(long);
    xbytes = sizeof(int) * num_tags;
    xbytes += xbytes % sizeof(long);
    total_bytes = (rbytes + xbytes * tnfa->num_states) * 2 + tbytes + pbytes;
    buf = alloca(total_bytes);
    if (buf == NULL)
      return REG_ESPACE;
    memset(buf, 0, total_bytes);
    tmp_tags = (void *)buf;
    tmp_buf = buf + tbytes;
    reach_next = (void *)tmp_buf;
    tmp_buf += rbytes;
    reach = (void *)tmp_buf;
    tmp_buf += rbytes;
    reach_pos = (void *)tmp_buf;
    tmp_buf += pbytes;
    for (i = 0; i < tnfa->num_states; i++)
      {
	reach[i].tags = (void *)tmp_buf;
	tmp_buf += xbytes;
	reach_next[i].tags = (void *)tmp_buf;
	tmp_buf += xbytes;
      }
  }

  for (i = 0; i < tnfa->num_states; i++)
    reach_pos[i].pos = -1;

  reach_next_i = reach_next;
  pos = 0;
  GET_NEXT_WCHAR();

#if 1
  /* If only one character can start a match, find it first. */
  if (tnfa->first_char >= 0)
    {
      const char *orig_str = str_byte;
      int first = tnfa->first_char;
      /* This uses strchr() hoping it might be optimized for the
	 target architecture. */
      /* XXX - use memchr(), wcschr() and wmemchr() as well! */
      str_byte = strchr(orig_str, first);
      if (str_byte == NULL)
	return REG_NOMATCH;
      prev_c = *(str_byte - 2);
      next_c = *(str_byte - 1);
      pos += str_byte - orig_str;
      DPRINT(("skipped %d chars\n", str_byte - orig_str));
    }
#endif

#if 0
  /* Skip over characters that cannot possibly be the first character
     of a match. */
  if (tnfa->firstpos_chars != NULL)
    {
      char *chars = tnfa->firstpos_chars;

      if (len < 0)
	{
	  const char *orig_str = str_byte;
	  /* XXX - use strpbrk() and wcspbrk() because they might be
	     optimized for the target architecture.  Try also strcspn()
	     and wcscspn() and compare the speeds. */
	  while (next_c != L'\0' && !chars[next_c])
	    {
	      next_c = *str_byte++;
	    }
	  prev_c = *(str_byte - 2);
	  pos += str_byte - orig_str;
	  DPRINT(("skipped %d chars\n", str_byte - orig_str));
	}
      else
	{
	  while (pos <= len && !chars[next_c])
	    {
	      prev_c = next_c;
	      next_c = (unsigned char)(*str_byte++);
	      pos++;
	    }
	}
    }
#endif

  DPRINT(("length: %d\n", len));
  DPRINT(("pos:chr/code | states and tags\n"));
  DPRINT(("-------------+------------------------------------------------\n"));

  while (1)
    {
      /* If no match found yet, add the initial states to `reach_next'. */
      if (match_eo < 0)
	{
	  DPRINT((" init >"));
	  trans_i = tnfa->initial;
	  while (trans_i->state != NULL)
	    {
	      if (reach_pos[trans_i->state_id].pos < pos)
		{
		  if (trans_i->assertions
		      && CHECK_ASSERTIONS(trans_i->assertions))
		    {
		      DPRINT(("assertion failed\n"));
		      trans_i++;
		      continue;
		    }

		  DPRINT((" %p", (void *)trans_i->state));
		  reach_next_i->state = trans_i->state;
		  for (i = 0; i < num_tags; i++)
		    reach_next_i->tags[i] = -1;
		  tag_i = trans_i->tags;
		  if (tag_i != NULL)
		    while (*tag_i >= 0)
		      {
			reach_next_i->tags[*tag_i] = pos;
			tag_i++;
		      }
		  if (reach_next_i->state == tnfa->final)
		    {
		      match_eo = pos;
		      for (i = 0; i < num_tags; i++)
			match_tags[i] = reach_next_i->tags[i];
		    }
		  reach_pos[trans_i->state_id].pos = pos;
		  reach_pos[trans_i->state_id].tags = &reach_next_i->tags;
		  reach_next_i++;
		}
	      trans_i++;
	    }
	  DPRINT(("\n"));
	  reach_next_i->state = NULL;
	}
      else
	{
	  if (num_tags == 0 || reach_next_i == reach_next)
	    /*We have found a match. */
	    break;
	}

      GET_NEXT_WCHAR();

#ifdef TRE_DEBUG
      DPRINT(("%3d:%2lc/%05d |", pos, (tre_cint_t)prev_c, (int)prev_c));
      print_reach(tnfa, reach_next);
#endif /* TRE_DEBUG */

      /* Check for end of string. */
      if (len < 0)
	{
	  if (prev_c == L'\0')
	    break;
	}
      else
	{
	  if (pos + pos_add > len)
	    break;
	}

      /* Swap `reach' and `reach_next'. */
      reach_i = reach;
      reach = reach_next;
      reach_next = reach_i;

      pos += pos_add;

      /* For each state in `reach' see if there is a transition leaving with
	 the current input symbol to a state not yet in `reach_next', and
	 add the destination states to `reach_next'. */
      reach_i = reach;
      reach_next_i = reach_next;
      while (reach_i->state != NULL)
	{
	  trans_i = reach_i->state;
	  while (trans_i->state != NULL)
	    {
	      /* Does this transition match the input symbol? */
	      if (trans_i->code_min <= prev_c &&
		  trans_i->code_max >= prev_c)
		{
		  if (trans_i->assertions
		      && (CHECK_ASSERTIONS(trans_i->assertions)
			  /* Handle character class transitions. */
			  || ((trans_i->assertions & ASSERT_CHAR_CLASS)
			      && !(cflags & REG_ICASE)
			      && !tre_isctype((tre_cint_t)prev_c, trans_i->class))
			  || ((trans_i->assertions & ASSERT_CHAR_CLASS)
			      && (cflags & REG_ICASE)
			      && (!tre_isctype(tre_tolower((tre_cint_t)prev_c),
					    trans_i->class)
				  && !tre_isctype(tre_toupper((tre_cint_t)prev_c),
					       trans_i->class)))
			  || ((trans_i->assertions & ASSERT_CHAR_CLASS_NEG)
			      && neg_char_classes_match(trans_i->neg_classes,
							(tre_cint_t)prev_c,
							cflags & REG_ICASE))))
		    {
		      DPRINT(("assertion failed\n"));
		      trans_i++;
		      continue;
		    }

		  /* Compute the tags after this transition. */
		  for (i = 0; i < num_tags; i++)
		    tmp_tags[i] = reach_i->tags[i];
		  tag_i = trans_i->tags;
		  if (tag_i != NULL)
		    while (*tag_i >= 0)
		      {
			tmp_tags[*tag_i] = pos;
			tag_i++;
		      }

		  if (reach_pos[trans_i->state_id].pos < pos)
		    {
		      /* Found an unvisited node. */
		      reach_next_i->state = trans_i->state;
		      tmp_iptr = reach_next_i->tags;
		      reach_next_i->tags = tmp_tags;
		      tmp_tags = tmp_iptr;
		      reach_pos[trans_i->state_id].pos = pos;
		      reach_pos[trans_i->state_id].tags = &reach_next_i->tags;

		      if (reach_next_i->state == tnfa->final
			  && (match_eo == -1
			      || (num_tags > 0
				  && reach_next_i->tags[0] <= match_tags[0])))
			{
			  match_eo = pos;
			  for (i = 0; i < num_tags; i++)
			    match_tags[i] = reach_next_i->tags[i];
			}
		      reach_next_i++;

		    }
		  else
		    {
		      assert(reach_pos[trans_i->state_id].pos == pos);
		      /* Another path has also reached this state.  We choose
			 the winner by examining the tag values for both
			 paths. */
		      if (tag_order(num_tags, tnfa->tag_directions, tmp_tags,
				    *reach_pos[trans_i->state_id].tags))
			{
			  /* The new path wins. */
			  tmp_iptr = *reach_pos[trans_i->state_id].tags;
			  *reach_pos[trans_i->state_id].tags = tmp_tags;
			  if (trans_i->state == tnfa->final)
			    {
			      match_eo = pos;
			      for (i = 0; i < num_tags; i++)
				match_tags[i] = tmp_tags[i];
			    }
			  tmp_tags = tmp_iptr;
			}
		    }
		}
	      trans_i++;
	    }
	  reach_i++;
	}
      reach_next_i->state = NULL;
    }

  DPRINT(("match end offset = %d\n", match_eo));

  *match_end_ofs = match_eo;
  return match_eo >= 0 ? REG_OK : REG_NOMATCH;
}

/* Fills the POSIX.2 regmatch_t array according to the TNFA tag and match
   endpoint values. */
static void
fill_pmatch(size_t nmatch, regmatch_t pmatch[],
	    const tre_tnfa_t *tnfa, int *tags, int match_eo)
{
  tre_submatch_data_t *submatch_data;
  int i;

  i = 0;
  if (match_eo >= 0 && !(tnfa->cflags & REG_NOSUB))
    {
      /* Construct submatch offsets from the tags. */
      DPRINT(("end tag = t%d = %d\n", tnfa->end_tag, match_eo));
      submatch_data = tnfa->submatch_data;
      while (i < tnfa->num_submatches && i < nmatch)
	{
	  if (submatch_data[i].so_tag == tnfa->end_tag)
	    pmatch[i].rm_so = match_eo - submatch_data[i].so_offset;
	  else
	    pmatch[i].rm_so =
	      tags[submatch_data[i].so_tag] - submatch_data[i].so_offset;

	  if (submatch_data[i].eo_tag == tnfa->end_tag)
	    pmatch[i].rm_eo = match_eo - submatch_data[i].eo_offset;
	  else
	    pmatch[i].rm_eo =
	      tags[submatch_data[i].eo_tag] - submatch_data[i].eo_offset;

	  /* If the boundary tag not set or is larger than the submatch
	     start, the submatch is reset. */
	  if (submatch_data[i].boundary >= 0
	      && (pmatch[i].rm_so < tags[submatch_data[i].boundary]
		  || tags[submatch_data[i].boundary] < 0))
	    {
	      DPRINT(("Submatch %d out of bounds (%d < %d), resetting\n",
		      i, pmatch[i].rm_so, tags[submatch_data[i].boundary]));
	      pmatch[i].rm_so = -1;
	      pmatch[i].rm_eo = -1;
	    }

	  if (submatch_data[i].marker_tag != 0)
	    {
	      int marker_tag, marker_boundary, marker_set;
	      marker_boundary = submatch_data[i].marker_boundary;
	      marker_tag = submatch_data[i].marker_tag;
	      DPRINT(("marker_boundary = t%d = %d, marker_tag = t%d = %d, "
		      "offs = %d\n",
		      marker_boundary, tags[marker_boundary],
		      marker_tag, tags[marker_tag],
		      tnfa->marker_offs[marker_tag]));

	      /* We consider the marker to be set for this submatch if
		 - the marker tag is used, and
		   - there is no marker boundary tag, or
		   - the marker boundary tag is smaller than or equal to
		     marker tag minus the marker tag offset. */
	      marker_set =
		tags[marker_tag] >= 0
		&& (marker_boundary <= 0 || marker_boundary >= tnfa->num_tags
		    || (tags[marker_tag] -
			tnfa->marker_offs[marker_tag]
			>= tags[marker_boundary]));
	      if (marker_tag > 0 && !marker_set)
		{
		  DPRINT(("Marker %d for submatch %d is false, resetting.\n",
			  marker_tag, i));
		  pmatch[i].rm_so = -1;
		  pmatch[i].rm_eo = -1;
		}
	    }

	  DPRINT(("pmatch[%d] = {t%d - %d = %d, t%d - %d = %d}\n", i,
		  submatch_data[i].so_tag, submatch_data[i].so_offset,
		  pmatch[i].rm_so,
		  submatch_data[i].eo_tag, submatch_data[i].eo_offset,
		  pmatch[i].rm_eo));

	  i++;
	}
    }
  while (i < nmatch)
    {
      pmatch[i].rm_so = -1;
      pmatch[i].rm_eo = -1;
      i++;
    }
}

static int
tnfa_match(const tre_tnfa_t *tnfa, const void *string, size_t len,
	   tre_str_type_t type, size_t nmatch, regmatch_t pmatch[],
	   int eflags)
{
  reg_errcode_t status;
  int *tags = NULL, eo;
  if (nmatch == 0)
    eflags |= REG_NOTAGS;
  if (tnfa->num_tags > 0 && !(eflags & REG_NOTAGS))
    {
      tags = alloca(sizeof(*tags) * tnfa->num_tags);
      if (tags == NULL)
	return REG_ESPACE;
    }
  status = tnfa_execute(tnfa, string, len, type, tags, eflags, &eo);
  if (status == REG_OK)
    fill_pmatch(nmatch, pmatch, tnfa, tags, eo);
  return status;
}

int
regnexec(const regex_t *preg, const char *str, size_t len,
	 size_t nmatch, regmatch_t pmatch[], int eflags)
{
#ifdef TRE_MULTIBYTE
  if (MB_CUR_MAX == 1)
    return tnfa_match(preg->value, str, len, STR_BYTE, nmatch, pmatch, eflags);
  else
    return tnfa_match(preg->value, str, len, STR_MBS, nmatch, pmatch, eflags);
#else /* !TRE_MULTIBYTE */
  return tnfa_match(preg->value, str, len, STR_BYTE, nmatch, pmatch, eflags);
#endif /* !TRE_MULTIBYTE */
}

int
regexec(const regex_t *preg, const char *str,
	size_t nmatch, regmatch_t pmatch[], int eflags)
{
  return regnexec(preg, str, -1, nmatch, pmatch, eflags);
}


#ifdef TRE_WCHAR

int
regwnexec(const regex_t *preg, const wchar_t *str, size_t len,
	  size_t nmatch, regmatch_t pmatch[], int eflags)
{
  return tnfa_match(preg->value, str, len, STR_WIDE, nmatch, pmatch, eflags);
}

int
regwexec(const regex_t *preg, const wchar_t *str,
	 size_t nmatch, regmatch_t pmatch[], int eflags)
{
  return regwnexec(preg, str, -1, nmatch, pmatch, eflags);
}

#endif /* TRE_WCHAR */

/*
   Approximate matching code.
*/

#ifdef TRE_APPROX

typedef struct {
  tre_tnfa_transition_t *state;
  int cost;
  int state_id;  /* XXX - this could be removed by using tre_tnfa_state_t
		    for `state'. */
} tre_tnfa_approx_reach_t;

typedef struct {
  int pos;
  int cost;
  tre_tnfa_approx_reach_t *reach;
} tre_approx_reach_pos_t;


static reg_errcode_t
tnfa_execute_approx(const tre_tnfa_t *tnfa, const void *string, size_t len,
		    tre_str_type_t type, int *match_tags,
		    int cost_ins, int cost_del, int cost_subst, int max_cost,
		    int eflags, int *match_end_ofs)
{
  tre_char_t prev_c = 0, next_c = 0;
  const char *str_byte = string;
  char *buf;
  int pos, pos_add = 1;
  tre_tnfa_transition_t *trans_i;
  tre_tnfa_approx_reach_t *reach, *reach_next, *reach_i, *reach_next_i;
  tre_approx_reach_pos_t *reach_pos;
  int i;
  int reg_notbol = eflags & REG_NOTBOL;
  int reg_noteol = eflags & REG_NOTEOL;
  int reg_newline = tnfa->cflags & REG_NEWLINE;
  int match_eo = -1;       /* end offset of match (-1 if no match found yet) */
  int match_cost = -1;
  int cflags = tnfa->cflags;
#ifdef TRE_WCHAR
  const wchar_t *str_wide = string;
#ifdef TRE_MULTIBYTE
  int pos_add_next = 0;
#ifdef HAVE_MBSTATE_T
  mbstate_t mbstate;
  memset(&mbstate, '\0', sizeof(mbstate));
#endif /* HAVE_MBSTATE_T */
#endif /* TRE_MULTIBYTE */
#endif /* !TRE_WCHAR */

  DPRINT(("tnfa_execute_approx, input type %d\n", type));

  /* Allocate memory for temporary data required for matching.  This needs to
     be done for every matching operation to be thread safe.  This allocates
     everything in a single large block from the stack frame using alloca(). */
  {
    int rbytes, pbytes, total_bytes;
    char *tmp_buf;
    rbytes = sizeof(*reach_next) * (tnfa->num_states + 1);
    rbytes += rbytes % sizeof(long);
    pbytes = sizeof(*reach_pos) * tnfa->num_states;
    pbytes += pbytes % sizeof(long);
    total_bytes = rbytes * 2 + pbytes;
    buf = alloca(total_bytes);
    if (buf == NULL)
      return REG_ESPACE;
    memset(buf, 0, total_bytes);
    reach_next = (void *)buf;
    tmp_buf = buf + rbytes;
    reach = (void *)tmp_buf;
    tmp_buf += rbytes;
    reach_pos = (void *)tmp_buf;
    tmp_buf += pbytes;
    for (i = 0; i < tnfa->num_states; i++)
      {
	reach[i].cost = 0;
	reach_next[i].cost = 0;
      }
  }

  for (i = 0; i < tnfa->num_states; i++)
    reach_pos[i].pos = -1;

  reach_next_i = reach_next;
  pos = 0;
  GET_NEXT_WCHAR();

  DPRINT(("length: %d\n", len));
  DPRINT(("pos:chr/code | states and costs\n"));
  DPRINT(("-------------+------------------------------------------------\n"));

  while (1)
    {
      /* If no match found yet, add the initial states to `reach_pos'. */
      if (match_eo < 0)
	{
	  DPRINT((" init >"));
	  trans_i = tnfa->initial;
	  while (trans_i->state != NULL)
	    {
	      if (reach_pos[trans_i->state_id].pos < pos)
		{
		  if (trans_i->assertions
		      && CHECK_ASSERTIONS(trans_i->assertions))
		    {
		      DPRINT(("assertion failed\n"));
		      trans_i++;
		      continue;
		    }

		  DPRINT((" %p", (void *)trans_i->state));
		  reach_next_i->state = trans_i->state;
		  if (reach_next_i->state == tnfa->final)
		    match_eo = pos;
		  reach_next_i->cost = 0;
		  reach_next_i->state_id = trans_i->state_id;
		  reach_pos[trans_i->state_id].pos = pos;
		  reach_pos[trans_i->state_id].cost = 0;
		  reach_pos[trans_i->state_id].reach = reach_next_i;
		  reach_next_i++;
		}
	      trans_i++;
	    }
	  DPRINT(("\n"));

	  /* Handle insertions -- this is done by pretending there is an
	     epsilon transition from each state back to the same state
	     with cost `cost_ins'. */
	  reach_i = reach;
	  while (reach_i->state != NULL)
	    {
	      int cost = reach_i->cost + cost_ins;
	      if (cost <= max_cost && reach_pos[reach_i->state_id].pos < pos)
		{
		  DPRINT(("insertion: state %d, cost %d\n",
			  reach_i->state_id, cost));
		  reach_next_i->cost = cost;
		  reach_next_i->state = reach_i->state;
		  reach_next_i->state_id = reach_i->state_id;
		  reach_pos[reach_i->state_id].pos = pos;
		  reach_pos[reach_i->state_id].cost = cost;
		  reach_pos[reach_i->state_id].reach = reach_next_i;
		  reach_next_i++;
		}
	      reach_i++;
	    }
	}
      else
	{
	  if (reach_next_i == reach_next)
	    /*We have found a match. */
	    break;
	}


      /* Handle deletions -- this is done by traversing through the whole
	 TNFA pretending that all transitions are epsilon transitions with
	 cost `cost_del'. */
      {
	/* XXX - how big should the ring buffer be? */
	tre_tnfa_approx_reach_t ringbuffer[512];
	tre_tnfa_approx_reach_t *deque_start, *deque_end;

	deque_start = deque_end = ringbuffer;
	reach_i = reach_next;
	/* Add all states in `reach_next' to the deque. */
	while (reach_i->state != NULL)
	  {
	    deque_end->state = reach_i->state;
	    deque_end->cost = reach_i->cost;
	    deque_end++;
	    reach_i++;
	  }

	/* Repeat until the deque is empty */
	while (deque_end != deque_start)
	  {
	    tre_tnfa_transition_t *trans_i;
	    tre_tnfa_approx_reach_t *reach_p;
	    int cost;

	    /* Pop the first state off the deque. */
	    reach_p = deque_start;
	    cost = reach_p->cost + cost_del;
	    if (cost > max_cost)
	      {
		deque_start++;
		if (deque_start >= (ringbuffer + 512))
		  deque_start = ringbuffer;
		continue;
	      }
	    trans_i = reach_p->state;

	    while (trans_i->state != NULL)
	      {
		/* XXX - should use input order counts to optimize! */
		/* XXX - should we check some assertions (BOL, EOL)? */
		if (reach_pos[trans_i->state_id].pos < pos
		    || reach_pos[trans_i->state_id].cost > cost)
		  {
		    DPRINT(("closure: state %d, cost %d\n", trans_i->state_id,
			    cost));
		    if (reach_pos[trans_i->state_id].pos < pos)
		      {
			/* Not reached yet. */
			reach_next_i->state = trans_i->state;
			reach_next_i->state_id = trans_i->state_id;
			reach_next_i->cost = cost;
			reach_next_i++;
		      }
		    else
		      {
			/* Already reached.  Reset cost. */
			reach_pos[trans_i->state_id].reach->cost = cost;
			reach_pos[trans_i->state_id].reach->state =
			  trans_i->state;
		      }
		    reach_pos[trans_i->state_id].cost = cost;
		    reach_pos[trans_i->state_id].pos = pos;

		    /* Add to the end of the deque. */
		    deque_end->state = trans_i->state;
		    deque_end->cost = cost;
		    deque_end++;
		    if (deque_end >= (ringbuffer + 512))
		      deque_end = ringbuffer;
		    assert(deque_end != deque_start);
		  }
		trans_i++;
	      }

	    deque_start++;
	    if (deque_start >= (ringbuffer + 512))
	      deque_start = ringbuffer;
	  }
      }

      reach_next_i->state = NULL;

      GET_NEXT_WCHAR();

#ifdef TRE_DEBUG
      DPRINT(("%3d:%2lc/%05d |", pos, (tre_cint_t)prev_c, (int)prev_c));
      reach_i = reach_next;
      while (reach_i->state != NULL)
	{
	  DPRINT((" %p:%d", (void *)reach_i->state, reach_i->cost));
	  reach_i++;
	}
      DPRINT(("\n"));
#endif /* TRE_DEBUG */

      /* Check for end of string. */
      if (len < 0)
	{
	  if (prev_c == L'\0')
	    break;
	}
      else
	{
	  if (pos + pos_add > len)
	    break;
	}

      /* Swap `reach' and `reach_next'. */
      reach_i = reach;
      reach = reach_next;
      reach_next = reach_i;

      pos += pos_add;

      /* Go through each state in `reach' and the transitions to
	 other states. */
      reach_i = reach;
      reach_next_i = reach_next;
      while (reach_i->state != NULL)
	{
	  trans_i = reach_i->state;
	  while (trans_i->state != NULL)
	    {
	      int cost = reach_i->cost;
	      if (trans_i->code_min > prev_c ||
		  trans_i->code_max < prev_c)
		{
		  /* The required character was not in the string.  Increase
		     cost by `cost_subst'. */
		  cost += cost_subst;
		  if (cost > max_cost)
		    {
		      trans_i++;
		      continue;
		    }
		}

	      if (trans_i->assertions
		  && (CHECK_ASSERTIONS(trans_i->assertions)
		      /* Handle character class transitions. */
		      || ((trans_i->assertions & ASSERT_CHAR_CLASS)
			  && !(cflags & REG_ICASE)
			  && !tre_isctype((tre_cint_t)prev_c, trans_i->class))
		      || ((trans_i->assertions & ASSERT_CHAR_CLASS)
			  && (cflags & REG_ICASE)
			  && (!tre_isctype(tre_tolower((tre_cint_t)prev_c),
					trans_i->class)
			      && !tre_isctype(tre_toupper((tre_cint_t)prev_c),
					   trans_i->class)))
		      || ((trans_i->assertions & ASSERT_CHAR_CLASS_NEG)
			  && neg_char_classes_match(trans_i->neg_classes,
						    (tre_cint_t)prev_c,
						    cflags & REG_ICASE))))
		{
		  DPRINT(("assertion failed\n"));
		  trans_i++;
		  continue;
		}

	      if (reach_pos[trans_i->state_id].pos < pos)
		{
		  /* Found an unvisited node. */
		  reach_next_i->state = trans_i->state;
		  reach_next_i->state_id = trans_i->state_id;
		  reach_next_i->cost = cost;
		  reach_pos[trans_i->state_id].pos = pos;
		  reach_pos[trans_i->state_id].cost = cost;
		  reach_pos[trans_i->state_id].reach = reach_next_i;

		  if (reach_next_i->state == tnfa->final && match_eo == -1)
		    {
		      match_eo = pos;
		      match_cost = cost;
		    }
		  reach_next_i++;

		}
	      else
		{
		  assert(reach_pos[trans_i->state_id].pos == pos);
		  /* Another path has also reached this state.  We choose the
		     one with the smallest cost. */
		  if (reach_pos[trans_i->state_id].cost > cost)
		    {
		      /* The new path wins. */
		      reach_pos[trans_i->state_id].cost = cost;
		      reach_pos[trans_i->state_id].reach->cost = cost;
		      if (trans_i->state == tnfa->final)
			{
			  match_eo = pos;
			  match_cost = cost;
			}
		    }
		}
	      trans_i++;
	    }
	  reach_i++;
	}
      reach_next_i->state = NULL;
    }

  DPRINT(("match end offset = %d, match cost = %d\n", match_eo, match_cost));

  *match_end_ofs = match_eo;
  return match_eo >= 0 ? REG_OK : REG_NOMATCH;
}

static int
tnfa_match_approx(const tre_tnfa_t *tnfa, const void *string, size_t len,
		  tre_str_type_t type, size_t nmatch, regmatch_t pmatch[],
		  int cost_ins, int cost_del, int cost_subst, int max_cost,
		  int eflags)
{
  reg_errcode_t status;
  int *tags = NULL, eo;
  if (nmatch == 0 || 1)
    eflags |= REG_NOTAGS;
  if (tnfa->num_tags > 0 && !(eflags & REG_NOTAGS))
    {
      tags = alloca(sizeof(*tags) * tnfa->num_tags);
      if (tags == NULL)
	return REG_ESPACE;
    }
  status = tnfa_execute_approx(tnfa, string, len, type, tags, cost_ins,
			       cost_del, cost_subst, max_cost, eflags, &eo);
  if (status == REG_OK && 0)
    fill_pmatch(nmatch, pmatch, tnfa, tags, eo);
  return status;
}

int
reganexec(const regex_t *preg, const char *str, size_t len,
	  size_t nmatch, regmatch_t pmatch[],
	  int cost_ins, int cost_del, int cost_subst, int max_cost,
	  int eflags)
{
#ifdef TRE_MULTIBYTE
  if (MB_CUR_MAX == 1)
    return tnfa_match_approx(preg->value, str, len, STR_BYTE, nmatch, pmatch,
			     cost_ins, cost_del, cost_subst, max_cost, eflags);
  else
    return tnfa_match_approx(preg->value, str, len, STR_MBS, nmatch, pmatch,
			     cost_ins, cost_del, cost_subst, max_cost, eflags);
#else /* !TRE_MULTIBYTE */
  return tnfa_match_approx(preg->value, str, len, STR_BYTE, nmatch, pmatch,
			   cost_ins, cost_del, cost_subst, max_cost, eflags);
#endif /* !TRE_MULTIBYTE */
}

int
regaexec(const regex_t *preg, const char *str,
	 size_t nmatch, regmatch_t pmatch[],
	 int cost_ins, int cost_del, int cost_subst, int max_cost,
	 int eflags)
{
  return reganexec(preg, str, -1, nmatch, pmatch, cost_ins, cost_del,
		   cost_subst, max_cost, eflags);
}

int
regawnexec(const regex_t *preg, const wchar_t *str, size_t len,
	   size_t nmatch, regmatch_t pmatch[],
	   int cost_ins, int cost_del, int cost_subst, int max_cost,
	   int eflags)
{
  return tnfa_match_approx(preg->value, str, len, STR_WIDE, nmatch, pmatch,
			   cost_ins, cost_del, cost_subst, max_cost, eflags);
}

int
regawexec(const regex_t *preg, const wchar_t *str,
	  size_t nmatch, regmatch_t pmatch[],
	  int cost_ins, int cost_del, int cost_subst, int max_cost,
	  int eflags)
{
  return regawnexec(preg, str, -1, nmatch, pmatch,
		    cost_ins, cost_del, cost_subst, max_cost, eflags);
}

#endif /* TRE_APPROX */

/* EOF */
