/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef ABSTRACTRESPONSEFILTER_H_
#define ABSTRACTRESPONSEFILTER_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include "httpheaderscollection.h"
#include "httpresponsemetadata.h"

using namespace std;

/*
The rules for AbstractResponseFilter implementations:

1. The order in which the caller calls the consume*() methods is as follows:
     consumeMetadata() [once]
     consumeDataChunk() [any number of times including zero]
     consumeEOF() [once]
   The consumeDataChunk() or the consumeEOF() methods won't be called until after the
   consumeMetadata() method has been called.
2. It is possible that none of the consume*() methods will be called. If the consumeMetadata()
   method has been called, the filter should not rely on the consumeEOF() to be called later.
   The situation when consumeEOF() is not being called may occur because of the server closing
   connection, or because the incorrect behaviour of other filters in a chain.
3. When the filter is done with modifying the metadata, it must start returning true on
   the isMetadataReady() call. Once a true is returned, it must be returned for all of
   the subsquent calls. After that, the filter must not modify the metadata.
4. If a filter always returns false on the isMetadataReadyCall(), then the response won't
   be sent out at all. This feature is used by the BlackHoleResponseFilter which is used
   to filter out 1xx responses when necessary.
5. After consumeEOF() has been called, the filter must always return true on the
   isMetadataReady() and isEOF() calls, unless it's not going to let the response out.
   A filter may return true on the isEOF() call before it got the consumeEOF() call
   (but after consumeMetadata()), if it has finished producing the data and is willing to
   ignore any further consumeDataChunk() calls. A filter can't return true on the isEOF() call
   but false on the isMetadataReady() call.
6. The filters that modify the body of the message, must use the setBodyStatus() and
   setBodySize() methods of the metadata object according to their actions. This must be done
   before the filter starts returning true on the isMetadataReady() call.
*/

class AbstractResponseFilter
{
public:
	virtual ~AbstractResponseFilter() {}
	virtual void consumeMetadata(HttpResponseMetadata& metadata) = 0;
	virtual void consumeDataChunk(const string& data) = 0;
	virtual void consumeEOF() = 0;
	virtual bool isMetadataReady() const = 0;
	virtual bool hasDataAvailable() const = 0;
	virtual void readDataChunk(string& collector) = 0;
	virtual bool isEOF() const = 0;
};

#endif

