/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "clientrequeststream.h"

#include <algorithm>
#include "bfilter.h"
#include "httperrormessages.h"
#include "conf.h"
#include "utils.h"
#include "debug.h"

ClientRequestStream::~ClientRequestStream()
{
	if (m_stream_writer) {
		delete m_stream_writer;
	}
}

void
ClientRequestStream::eventError(int ecode, const char *info)
{
	if (!info) {
		info = "";
	}
	DEBUGLOG("ClientRequestStream::eventError(ecode = " << ecode << ")");
	
	HttpClientRequest *last_request = m_owner.requestsToBeServed().empty() ? 0 : m_owner.requestsToBeServed().back();
	HttpClientRequest *request = last_request;
	
	if (!last_request || last_request->isComplete()) {
		// the request which generated this error, has not yet emerged
		request = new HttpClientRequest(
			HttpRequestMetadata(
				HttpRequestLine(
					"GET", URI("http://dummyhost/bf.intercept-request/error_substitute"),
					HttpVersion::HTTP_1_0
				)
			)
		);
		m_owner.requestsToBeServed().push_back(request);
	}
	
	string error_message;
	switch (ecode) {
	case ERR_PARSING_REQUEST_LINE:
		error_message = HttpErrorMessages::errMalformedRequestLine(request->getOrigRequestLine());
		break;
	case ERR_PARSING_HEADER:
		error_message = HttpErrorMessages::errMalformedRequestHeader(request->getOrigRequestLine());
		break;
	case ERR_PARSING_CHUNK_HEADER:
		error_message = HttpErrorMessages::errMalformedChunkInRequest(request->getOrigRequestLine());
		break;
	case ERR_LENGTH_REQUIRED:
		error_message = HttpErrorMessages::errRequestBodyLengthRequired(request->getOrigRequestLine());
		break;
	case ERR_UNEXPECTED_EOF:
		m_owner.finishSession();
		break;
	case ERR2_BAD_REQUEST_URI:
		error_message = HttpErrorMessages::errBadRequestURI(request->getOrigRequestLine(), info);
		break;
	case ERR2_UNSUPPORTED_METHOD:
		error_message = HttpErrorMessages::errUnsupportedRequestMethod(request->getOrigRequestLine());
		break;
	case ERR2_UNSUPPORTED_EXPECTATION:
		error_message = HttpErrorMessages::errUnsupportedExpectation(request->getOrigRequestLine(), info);
		break;
	default:
		error_message = HttpErrorMessages::errInternal(request->getOrigRequestLine(),
			"Unhandled error in ClientRequestStream: "+num2str(ecode));
	}
	
	request->setInterceptHandler(HttpClientRequest::IH_ERROR_SUBSTITUTE);
	request->setErrorResponse(error_message);
	request->setComplete();
	if (m_owner.isRequestSentToServer(request)) {
		m_owner.requestsSentToServer().pop_back();
	}
	dispatchMoreRequests();
}

void
ClientRequestStream::eventGotMetadata(const HttpRequestMetadata& metadata)
{
	if (metadata.requestLine().getHttpVersion() > m_owner.clientInfo().getProtocolCompability()) {
		m_owner.clientInfo().setProtocolCompability(metadata.requestLine().getHttpVersion());
	}
	
	if (m_stream_writer) {
		DEBUGLOG2("BUG");
		delete m_stream_writer;
		m_stream_writer = 0;
	}
	
	DEBUGLOG("Request begins: " << metadata.requestLine().getURI().toString());
	
	HttpClientRequest *request = new HttpClientRequest(metadata);
	prepareRequestForForwarding(request->metadata());
	m_owner.requestsToBeServed().push_back(request);
	
	string method = metadata.requestLine().getMethod();
	if (method != "GET" && method != "POST" && method != "PUT" && method != "HEAD") {
		eventError(ERR2_UNSUPPORTED_METHOD, method.c_str());
	}
	if (!metadata.requestLine().getURI().isAbsolute()) {
		eventError(ERR2_BAD_REQUEST_URI, "URI must be absolute");
	} else if (metadata.requestLine().getURI().getScheme() != "http") {
		eventError(ERR2_BAD_REQUEST_URI, "Scheme not supported");
	} else {
		list<HttpHeaderElement> elements = metadata.headers().getHeader("Expect").elements().getElements();
		list<HttpHeaderElement>::const_iterator iter = elements.begin();
		for (; iter != elements.end(); ++iter) {
			if (iter->getLowercaseName() != "100-continue") {
				eventError(ERR2_UNSUPPORTED_EXPECTATION, iter->getName().c_str());
				break;
			}
		}
	}
	
	dispatchMoreRequests();
}

void
ClientRequestStream::eventGotContentPart(const string& data)
{
	(*m_owner.requestsToBeServed().rbegin())->appendBodyData(data);
	if (m_stream_writer) {
		m_stream_writer->write(data);
	}
}

void
ClientRequestStream::eventMessageEnds(bool connection_needs_closing)
{
	DEBUGLOG("ClientRequestStream::eventMessageEnds()");
	m_owner.serverInfo().registerRequest(); // FIXME: this does not mean the request has been sent
#ifdef DEBUG
	if (m_owner.requestsToBeServed().empty()) {
		DEBUGLOG2("BUG");
		m_owner.finishSession();
	}
#endif
	HttpClientRequest *request = *m_owner.requestsToBeServed().rbegin();
#ifdef DEBUG
	if (connection_needs_closing) {
		DEBUGLOG2("BUG");
	}
#endif
	request->setComplete();
	if (m_stream_writer) {
		m_stream_writer->end();
		delete m_stream_writer;
		m_stream_writer = 0;
		if (request->isAlreadyServed()) {
			delete request;
			m_owner.requestsToBeServed().pop_back();
#ifdef DEBUG
			if (!m_owner.requestsToBeServed().empty()) {
				DEBUGLOG2("BUG");
			}
#endif
		}
	}
	
	DEBUGLOG("Request ends (requests_sent_to_server: " << m_owner.requestsSentToServer().size()
		<< ", requests_to_be_served: " << m_owner.requestsToBeServed().size() << ")");
	
	if (m_owner.requestsSentToServer().empty()) {
		dispatchMoreRequests();
	}
}

void
ClientRequestStream::moreData(const string& data)
{
	m_owner.dataForServer(data);
}

void
ClientRequestStream::done(bool conn_close)
{
#ifdef DEBUG
	if (conn_close) {
		DEBUGLOG2("BUG");
	}
#endif
}

void
ClientRequestStream::prepareRequestForForwarding(HttpRequestMetadata& metadata)
{
	HttpHeader connection_hdr = metadata.headers().getHeader("Connection");
	connection_hdr.elements().appendElements(metadata.headers().getHeader("Proxy-Connection").elements());
	list<HttpHeaderElement> hop_by_hop_headers = connection_hdr.elements().getElements();
	list<HttpHeaderElement>::iterator iter = hop_by_hop_headers.begin();
	for (; iter != hop_by_hop_headers.end(); ++iter) {
		metadata.headers().removeHeader(iter->getName());
	}
	metadata.requestLine().setHttpVersion(HttpVersion::HTTP_1_1);
	metadata.headers().removeHeader("Keep-Alive");
	metadata.headers().removeHeader("Proxy-Connection");
	metadata.headers().removeHeader("Upgrade");
	metadata.headers().setHeader(HttpHeader("Connection", "te"));
	metadata.headers().setHeader(HttpHeader("Accept-Encoding", "gzip, deflate"));
	metadata.headers().setHeader(HttpHeader("TE", "gzip, deflate"));
	if (!metadata.headers().hasHeader("Host")) {
		string host = metadata.requestLine().getURI().getHost();
		int port = metadata.requestLine().getURI().getPort();
		if (port >= 0) {
			host += ":"+num2str(port);
		}
		metadata.headers().setHeader(HttpHeader("Host", host));
	}
	if (!Conf::getInstance().usingProxy()) {
		URI uri = metadata.requestLine().getURI();
		uri.makeRelative();
		metadata.requestLine().setURI(uri);
	}
}

void
ClientRequestStream::dispatchMoreRequests()
{
	DEBUGLOG("dispatchMoreRequests()");
	
	list<HttpClientRequest *>::iterator iter = m_owner.requestsToBeServed().begin();
	for (; iter != m_owner.requestsToBeServed().end(); ++iter) {
		if (!m_owner.isRequestSentToServer(*iter)) {
			break;
		}
	}
	
	while (iter != m_owner.requestsToBeServed().end()) {
		HttpClientRequest *request = *iter;
		
		if (!request->isToBeSent()) {
			if (!request->isComplete()) {
				return; // wait for it to finish
			}
			if (m_owner.requestsSentToServer().empty()) {
				m_owner.responseStream().processSubstitutionRequest(request);
				delete request;
				m_owner.requestsToBeServed().erase(iter++);
				DEBUGLOG("Substitution request served (requests_sent_to_server: " << m_owner.requestsSentToServer().size()
					<< ", requests_to_be_served: " << m_owner.requestsToBeServed().size() << ")");
				continue;
			}
			return; // first the server must serve the requests we've sent to it
		}
		
		HttpClientRequest *last_request_sent_to_server = m_owner.requestsSentToServer().empty() ? 0
			: *m_owner.requestsSentToServer().rbegin();
		bool send_it_now = true;
		if (last_request_sent_to_server) {
			if (!last_request_sent_to_server->isIdempotent()) {
				send_it_now = false;
			} else if (!isSameServer(request->getAbsoluteURI(), last_request_sent_to_server->getAbsoluteURI())) {
				send_it_now = false;
			} else if (m_owner.serverInfo().getPipeliningCapability() == ServerInfo::P_UNKNOWN
			           && m_owner.serverInfo().getServedRequestsCount() > 0) {
				m_owner.serverInfo().setPipeliningCapability(ServerInfo::P_TESTING);
				// send_it_now = true;
				DEBUGLOG("Pipelining a single request to test the pipelining capability of the server");
			} else if (m_owner.serverInfo().getPipeliningCapability() != ServerInfo::P_OK) {
				send_it_now = false;
			}
		} else {
			m_owner.switchToServer(request->getAbsoluteURI().getHost(), request->getAbsoluteURI().guessPort());
		}
		if (!send_it_now) {
			return;
		}
		
		if (request->isComplete()) {
			m_owner.requestsSentToServer().push_back(request);
			HttpStreamWriter *stream_writer = new HttpStreamWriter(*this, request->metadata(), true, false);
			stream_writer->write(request->getBodyData());
			stream_writer->end();
			delete stream_writer;
			DEBUGLOG("Request sent to server (requests_sent_to_server: " << m_owner.requestsSentToServer().size()
				<< ", requests_to_be_served: " << m_owner.requestsToBeServed().size() << ")");
		} else {
#ifdef DEBUG
			if (iter != --m_owner.requestsToBeServed().end()) {
				DEBUGLOG2("BUG");
			}
#endif
			if (!request->isToBeSent() || request->isAlreadyServed()) {
				return; // need to wait for it to finish
			}
#ifdef DEBUG
			if (m_stream_writer) {
				DEBUGLOG2("BUG");
				m_owner.finishSession();
			}
#endif
			m_owner.requestsSentToServer().push_back(request);
			m_stream_writer = new HttpStreamWriter(*this, request->metadata(), true, false);
			m_stream_writer->write(request->getBodyData());
			DEBUGLOG("Sending incomplete request (requests_sent_to_server: " << m_owner.requestsSentToServer().size()
				<< ", requests_to_be_served: " << m_owner.requestsToBeServed().size() << ")");
			return;
		}
		++iter;
	}
}

bool
ClientRequestStream::isSameServer(const URI& one, const URI& other)
{
	return (strtolower(one.getHost()) == strtolower(other.getHost()) && one.guessPort() == other.guessPort());
}

