/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gifimageinspector.h"
#include <cstring>

const unsigned char GIFImageInspector::SIGNATURE[3] = { 'G', 'I', 'F' };

GIFImageInspector::GIFImageInspector()
:	m_state(ST_SIGNATURE), m_delay_state(DST_DELAY_NOT_ENCOUNTERED),
	m_inspection_in_progress(true), m_inspection_failed(false),
	m_width(0), m_height(0), m_is_animated(false), m_cur_pos(0),
	m_cur_block_length(0)
{}

void
GIFImageInspector::consumeData(const string& data)
{
	m_data += data;
	if (!m_inspection_in_progress) {
		return;
	}
	while (true) {
		if (m_state == ST_SIGNATURE) {
			if (m_data.length() >= sizeof(SIGNATURE)) {
				if (memcmp(m_data.data(), SIGNATURE, sizeof(SIGNATURE))) {
					m_inspection_in_progress = false;
					m_inspection_failed = true;
					break;
				} else {
					m_state = ST_HEADER;
				}
			} else {
				break;
			}
		}
		if (m_state == ST_HEADER) {
			if (m_data.length() >= 13) {
				const unsigned char *bytes = (const unsigned char *)m_data.data();
				m_width = (unsigned int)bytes[6]+((unsigned int)bytes[7]<<8);
				m_height = (unsigned int)bytes[8]+((unsigned int)bytes[9]<<8);
				unsigned char flag = bytes[10];
				if (flag & 0x80) {
					m_state = ST_GLOBAL_COLOR_TABLE;
					m_cur_block_length = 3*(1<<((flag&7)+1));
				} else {
					m_state = ST_BLOCK_BEGIN;
				}
				m_cur_pos = 13;
				if (!memcmp(bytes+3, "87a", 3)) {
					// this version can't be animated
					m_inspection_in_progress = false;
					return;
				}
			} else {
				break;
			}
		}
		if (m_state == ST_GLOBAL_COLOR_TABLE) {
			if (m_data.length() >= m_cur_pos+m_cur_block_length) {
				m_cur_pos += m_cur_block_length;
				m_state = ST_BLOCK_BEGIN;
			} else {
				break;
			}
		}
		if (m_state == ST_BLOCK_BEGIN) {
			if (m_data.length() >= m_cur_pos+1) {
				unsigned char block_id = m_data[m_cur_pos];
				switch (block_id) {
				case 0x2c:
					m_state = ST_IMAGE_DESCRIPTOR;
					if (m_delay_state == DST_WAITING_FOR_GRAPHIC_BLOCK) {
						m_delay_state = DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK;
					} else if (m_delay_state == DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK) {
						m_inspection_in_progress = false;
						m_is_animated = true;
						return;
					}
					break;
				case 0x21:
					m_state = ST_EXTENSION_BLOCK_BEGIN;
					break;
				case 0x3b:
					m_inspection_in_progress = false;
					return;
				default:
					m_inspection_in_progress = false;
					m_inspection_failed = true;
					return;
				}
			} else {
				break;
			}
		}
		if (m_state == ST_IMAGE_DESCRIPTOR) {
			if (m_data.length() >= m_cur_pos+10) {
				unsigned char flags = m_data[m_cur_pos+9];
				if (flags & 0x80) {
					m_state = ST_LOCAL_COLOR_TABLE;
					m_cur_block_length = 3*(1<<((flags&7)+1));
				} else {
					m_state = ST_IMAGE_DATA;
				}
				m_cur_pos += 10;
			} else {
				break;
			}
		}
		if (m_state == ST_LOCAL_COLOR_TABLE) {
			if (m_data.length() >= m_cur_pos+m_cur_block_length) {
				m_cur_pos += m_cur_block_length;
				m_state = ST_IMAGE_DATA;
			} else {
				break;
			}
		}
		if (m_state == ST_IMAGE_DATA) {
			if (m_data.length() >= m_cur_pos+1) {
				m_state = ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR;
				m_cur_pos += 1;
			} else {
				break;
			}
		}
		if (m_state == ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR) {
			if (m_data.length() >= m_cur_pos+1) {
				unsigned char byte = m_data[m_cur_pos];
				if (byte == 0x00) {
					m_state = ST_BLOCK_BEGIN;
				} else {
					m_state = ST_IMAGE_DATA_SUBBLOCK;
					m_cur_block_length = byte;
				}
				m_cur_pos += 1;
			} else {
				break;
			}
		}
		if (m_state == ST_IMAGE_DATA_SUBBLOCK) {
			if (m_data.length() >= m_cur_pos+m_cur_block_length) {
				m_cur_pos += m_cur_block_length;
				m_state = ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR;
			} else {
				break;
			}
		}
		if (m_state == ST_EXTENSION_BLOCK_BEGIN) {
			if (m_data.length() >= m_cur_pos+2) {
				unsigned char ext_type = m_data[m_cur_pos+1];
				if (ext_type == 0xf9) {
					m_state = ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR;
				} else {
					m_state = ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR;
					if (ext_type == 0x01) { // plain text extension
						if (m_delay_state == DST_WAITING_FOR_GRAPHIC_BLOCK) {
							m_delay_state = DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK;
						} else if (m_delay_state == DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK) {
							m_inspection_in_progress = false;
							m_is_animated = true;
							return;
						}
					}
				}
				m_cur_pos += 2;
			} else {
				break;
			}
		}
		if (m_state == ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR) {
			if (m_data.length() >= m_cur_pos+1) {
				unsigned char size = m_data[m_cur_pos];
				m_cur_pos += 1;
				if (size == 0x00) {
					m_state = ST_BLOCK_BEGIN;
				} else {
					m_state = ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK;
					m_cur_block_length = size;
					if (m_cur_block_length != 4) {
						m_inspection_in_progress = false;
						m_inspection_failed = true;
						return;
					}
				}
			} else {
				break;
			}
		}
		if (m_state == ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK) {
			if (m_data.length() >= m_cur_pos+m_cur_block_length) {
				if (m_data[m_cur_pos+1] != 0x00 || m_data[m_cur_pos+2] != 0x00) {
					if (m_delay_state == DST_DELAY_NOT_ENCOUNTERED) {
						m_delay_state = DST_WAITING_FOR_GRAPHIC_BLOCK;
					}
				}
				m_state = ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR;
				m_cur_pos += m_cur_block_length;
			} else {
				break;
			}
		}
		if (m_state == ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR) {
			if (m_data.length() >= m_cur_pos+1) {
				unsigned char size = m_data[m_cur_pos];
				m_cur_pos += 1;
				if (size == 0x00) {
					m_state = ST_BLOCK_BEGIN;
				} else {
					m_state = ST_OTHER_EXTENSION_SUBBLOCK;
					m_cur_block_length = size;
				}
			} else {
				break;
			}
		}
		if (m_state == ST_OTHER_EXTENSION_SUBBLOCK) {
			if (m_data.length() >= m_cur_pos+m_cur_block_length) {
				m_state = ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR;
				m_cur_pos += m_cur_block_length;
			} else {
				break;
			}
		}
	}
}

void
GIFImageInspector::consumeEOF()
{
	if (m_inspection_in_progress) {
		m_inspection_in_progress = false;
		m_inspection_failed = true;
	}
}

void
GIFImageInspector::reset()
{
	m_state = ST_SIGNATURE;
	m_delay_state = DST_DELAY_NOT_ENCOUNTERED;
	m_data.erase();
	m_inspection_in_progress = true;
	m_inspection_failed = false;
	m_width = 0;
	m_height = 0;
	m_is_animated = false;
	m_cur_pos = 0;
	m_cur_block_length = 0;
}

