/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstring>
#include "crc32.h"
#include "gzipcompressor.h"

GzipCompressor::GzipCompressor()
:	m_error(false), m_eof(false), m_stream_ready(false), m_output_flushed(false),
	m_unsynced_size(0), m_state(STATE_HEAD)
{}

GzipCompressor::~GzipCompressor()
{
	if (m_stream_ready)
		deflateEnd(&m_strm);
}

void
GzipCompressor::reset()
{
	if (m_stream_ready) {
		deflateEnd(&m_strm);
		m_stream_ready = false;
	}
	m_data.erase();
	m_error = false;
	m_eof = false;
	m_output_flushed = false;
	m_state = STATE_HEAD;
}

bool
GzipCompressor::readOutput(string& output)
{
	output.erase();
	if (m_error)
		return false;
	if (m_data.empty() && m_eof && m_output_flushed)
		return true;
	bool res = doReadOutput(output);
	m_error = !res || (m_eof && !m_data.empty());
	return res;
}

bool
GzipCompressor::doReadOutput(string& output)
{
	while (true) {
		switch (m_state) {
		case STATE_HEAD:
			unsigned char header[10];
			header[0] = 0x1f; // id #1
			header[1] = 0x8b; // id #2
			header[2] = 0x08; // compression method (0x08 means deflate)
			header[3] = 0x00; // flags
			header[4] = 0x00; // mtime0
			header[5] = 0x00; // mtime1
			header[6] = 0x00; // mtime2
			header[7] = 0x00; // mtime3
			header[8] = 0x00; // extra flags
			header[9] = 0xff; // os (0xff means unknown)
			output += string(header, header+sizeof(header));
			m_state = STATE_DATA;
			m_data_length = 0L;
			m_data_crc32 = 0L;
			break;
		case STATE_DATA:
			if (m_data.empty() && (!m_eof || m_output_flushed))
				return true;
			if (!m_stream_ready) {
				m_strm.zalloc = Z_NULL;
				m_strm.zfree = Z_NULL;
				m_strm.opaque = Z_NULL;
				m_strm.next_in = Z_NULL;
				deflateInit2(&m_strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED,
				             -MAX_WBITS, 8, Z_DEFAULT_STRATEGY);
				m_stream_ready = true;
			}
			int res;
			do {
				m_strm.next_in = m_inbuf;
				string::size_type indata_size = sizeof(m_inbuf)<m_data.length() ? sizeof(m_inbuf) : m_data.length();
				m_strm.avail_in = indata_size;
				m_unsynced_size += indata_size;
				memcpy(m_inbuf, m_data.c_str(), indata_size);
				do {
					m_strm.next_out = m_outbuf;
					m_strm.avail_out = sizeof(m_outbuf);
					m_output_flushed = m_eof;
					int flush = 0;
					if (m_eof && indata_size == m_data.length()) {
						flush = Z_FINISH;
						m_unsynced_size = 0;
					} else if (m_unsynced_size > 6000) {
						flush = Z_SYNC_FLUSH;
						m_unsynced_size = 0;
					}
					res = deflate(&m_strm, flush);
					output += string(m_outbuf, m_outbuf+sizeof(m_outbuf)-m_strm.avail_out);
				} while (res == Z_OK && m_strm.avail_out == 0);
				m_data_crc32 = CRC32::update(m_data.substr(0, indata_size-m_strm.avail_in), m_data_crc32);
				m_data_length += indata_size-m_strm.avail_in;
				m_data.erase(0, indata_size-m_strm.avail_in);
			} while (res == Z_OK && !m_data.empty());
			if (res == Z_STREAM_END) {
				if (m_stream_ready) {
					deflateEnd(&m_strm);
					m_stream_ready = false;
				}
				m_state = STATE_CRC32_ISIZE;
				break;
			} else if (m_output_flushed) {
				return false;
			}
			return (res == Z_OK);
			break;
		case STATE_CRC32_ISIZE:
			unsigned char tmp[8];
			tmp[0] = m_data_crc32 & 0xff;
			tmp[1] = (m_data_crc32 >> 8) & 0xff;
			tmp[2] = (m_data_crc32 >> 16) & 0xff;
			tmp[3] = (m_data_crc32 >> 24) & 0xff;
			tmp[4] = m_data_length & 0xff;
			tmp[5] = (m_data_length >> 8) & 0xff;
			tmp[6] = (m_data_length >> 16) & 0xff;
			tmp[7] = (m_data_length >> 24) & 0xff;
			output += string(tmp, tmp+sizeof(tmp));
			m_state = STATE_END;
			return true;
		default:
			return false;
		}
	}
	return false;
}

