/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <list>
#include <set>
#include <algorithm>
#include <iostream>
#include "bfilter.h"
#include "utils.h"
#include "htmlprocessor.h"
#include "rules.h"
#include "state.h"
#include "jsenvironment.h"
#include "htmltaginfo.h"
#include "httpfetcher.h"

using namespace std;

HTMLProcessor::HTMLProcessor(BFilter *owner)
:	m_is_subprocessor(false), m_owner(owner), m_content_type(CONTENT_HTML),
	m_base_url("http://dummy.url"),
	m_inside_head(false), m_onload_check_done(false),
	m_nest_level(0), m_webbug_open_comment(false)
{
	constructImportantTags();
}

HTMLProcessor::HTMLProcessor(BFilter *owner, const Context& context, JSEnvProxy& js_env, int nest_level)
:	m_is_subprocessor(true), m_owner(owner), m_content_type(context.m_content_type),
	m_base_url(context.m_base_url), m_inside_head(context.m_inside_head),
	m_script_3rd_party(context.m_script_3rd_party), m_onload_check_done(false),
	m_scripts(context.m_scripts), m_jsengine(context.m_jsengine),
	m_target_blank(context.m_target_blank), m_html_stack(context.m_html_stack),
	m_onload(context.m_onload), m_onunload(context.m_onunload),
	m_js_env(js_env), m_nest_level(nest_level), m_webbug_open_comment(false)
{
	constructImportantTags();
}

void
HTMLProcessor::setBaseURL(const URI& url)
{
	m_base_url = url;
	string url_str = url.toString();
	m_scripts = Rules::getInstance().getScriptsParam(url_str);
	m_jsengine = Rules::getInstance().getJSengineParam(url_str);
	m_target_blank = Rules::getInstance().getTargetBlankParam(url_str);
}

void
HTMLProcessor::consumeData(const string& data)
{
	m_lexer.consumeData(data);
	int tok;
	while ((tok = m_lexer.getNextToken())) {
		processToken(tok);
	}
}

void
HTMLProcessor::consumeEOF()
{
	m_lexer.consumeEOF();
	int tok;
	while ((tok = m_lexer.getNextToken())) {
		processToken(tok);
	}
	if (!m_cur_text.empty()) {
		m_html_stack.appendText(m_cur_text);
		m_cur_text.erase();
	}
	if (m_html_stack.bottom()) {
		m_processed_data += m_html_stack.toString(m_content_type==CONTENT_XHTML);
	}
	m_lexer.reset();
	m_closing_tag.erase();
	m_attr_name.erase();
	m_html_stack.clear();
	m_cur_tag = HTMLElementNode();
	m_inside_head = false;
	m_onload_check_done = false;
}

void
HTMLProcessor::reset()
{
	m_lexer.reset();
	m_processed_data.erase();
	m_closing_tag.erase();
	m_attr_name.erase();
	m_html_stack.clear();
	m_ignored_a_tags.clear();
	m_cur_tag = HTMLElementNode();
	m_cur_text.erase();
	m_inside_head = false;
	m_onload_check_done = false;
	if (!m_is_subprocessor) {
		m_js_env.reset();
	}
	m_onload.erase();
	m_onunload.erase();
}

void
HTMLProcessor::saveContext(Context& context) const
{
	context.m_content_type = m_content_type;
	context.m_base_url = m_base_url;
	context.m_inside_head = m_inside_head;
	context.m_script_3rd_party = m_script_3rd_party;
	context.m_scripts = m_scripts;
	context.m_jsengine = m_jsengine;
	context.m_target_blank = m_target_blank;
	context.m_html_stack = m_html_stack;
	if (!m_cur_text.empty()) {
		context.m_html_stack.appendText(m_cur_text);
	}
	context.m_onload = m_onload;
	context.m_onunload = m_onunload;
}

bool
HTMLProcessor::isExternalURL(const URI& url)
{
	string rhost = strtolower(url.getHost());
	if (rhost.find("www.") == 0)
		rhost.erase(0, 4);
	string lhost = strtolower(m_base_url.getHost());
	if (lhost.find("www.") == 0)
		lhost.erase(0, 4);
	if (rhost.empty()) {
		return false;
	}
	if (rhost == lhost) {
		return false;
	}
	// if rhost is a subdomain of lhost, return false
	string::size_type pos = rhost.rfind(lhost);
	if (pos != string::npos) {
		if (pos+lhost.length()==rhost.length()) {
			if (pos == 0 || rhost[pos-1] == '.') {
				return false;
			}
		}
	}
	return true;
}

bool
HTMLProcessor::isSuspiciousURL(const URI& url)
{
	//if (url.getQueryEncoded().length() > 35)
	//	return true;
	
	if (url.getDecodedPath().find("http://") != string::npos ||
	    url.getDecodedQuery().find("http://") != string::npos) {
		return true;
	}
	
	string path_plus_query = url.getRawPath()+"?"+url.getRawQuery();
	static string digits = "0123456789";
	string::size_type pos1 = 0;
	int ndigits = 0;
	while (true) {
		string::size_type pos2 = path_plus_query.find_first_of(digits, pos1);
		if (pos2 == string::npos)
			break;
		pos1 = pos2+1;
		++ndigits;
	}
	if (ndigits > 6)
		return true;
	
	pos1 = 0;
	int nequals = 0;
	while (true) {
		string::size_type pos2 = path_plus_query.find('=', pos1);
		if (pos2 == string::npos)
			break;
		pos1 = pos2+1;
		++nequals;
	}
	return (nequals > 3);
}

bool
HTMLProcessor::isCommonAdSize(int width, int height)
{
	if (width == 120 && height == 60) {
		return true;
	} else if (width == 120 && height == 600) {
		return true;
	} else if (width == 120 && height == 60) {
		return true;
	} else if (width == 125 && height == 125) {
		return true;
	} else if (width == 234 && height == 60) {
		return true;
	} else if (width == 468 && height == 60) {
		return true;
	} else if (width == 88 && height == 31) {
		return true;
	} else if (width == 728 && height == 90) {
		return true;
	} else if (width == 160 && height == 600) {
		return true;
	} else if (width == 300 && height == 250) {
		return true;
	} else if (width == 180 && height == 150) {
		return true;
	}
	return false;
}

bool
HTMLProcessor::couldBeAdSize(int width, int height)
{
	if (width == 0 || height == 0) {
		return false;
	}
	return (width * height >= 10000 && width * height < 120000 &&
		(float)std::max(width, height)/std::min(width, height) > 3.0);
}

int
HTMLProcessor::imageIsAd(int img_ad_rule, int width, int height, const string& target,
                         bool external_link, bool external_img, bool suspicious_link, bool suspicious_img,
                         int animated, int redir, int cookie) {
	enum {YES=1, NO=-1, NOT_SURE=0};
	if (img_ad_rule == 2) {
		return YES;
	}
	if (img_ad_rule == 1) {
		return NO;
	}
	if (width * height < 600) {
		// too small
		return NO;
	}
	if (width * height > 200000 && !external_img && !suspicious_img) {
		// too big
		return NO;
	}
	if (isCommonAdSize(width, height)) {
		if (external_link || external_img || suspicious_link || suspicious_img || redir == YES || animated == YES) {
			return YES;
		}
	} else {
		int factor = 0;
		if (couldBeAdSize(width, height)) {
			factor += 2;
		}
		if (external_link) factor++;
		if (suspicious_link) factor++;
		if (external_img) factor++;
		if (suspicious_img) factor++;
		factor += redir > 0 ? 2 : redir;
		factor += cookie;
		if (animated == YES) {
			factor++;
		}
		if (target == "_blank") {
			factor += 2;
		} else if (!target.empty()) {
			factor++;
		}
		if (factor > 4) {
			return YES;
		} else if (factor < 2) {
			return NO;
		}
	}
	return NOT_SURE;
}

void
HTMLProcessor::processToken(int tok)
{
	switch (tok) {
	case HTMLLexer::TOK_DOCTYPE:
		m_content_type = (m_lexer.tokText().find("XHTML") != string::npos) ? CONTENT_XHTML : CONTENT_HTML;
	case HTMLLexer::TOK_TEXT:
	case HTMLLexer::TOK_COMMENT:
	case HTMLLexer::TOK_CDATA:
		if (m_html_stack.top()) {
			m_cur_text += m_lexer.tokText();
		} else {
			m_processed_data += m_lexer.tokText();
		}
		break;
	case HTMLLexer::TOK_TAG_BEGIN:
		m_tag_explicit_empty = false;
		m_cur_tag = HTMLElementNode();
		if (!m_cur_text.empty()) {
			m_html_stack.appendText(m_cur_text);
			m_cur_text.erase();
		}
		break;
	case HTMLLexer::TOK_TAG_NAME:
		m_cur_tag.setName(strtolower(m_lexer.tokText()));
		break;
	case HTMLLexer::TOK_TAG_EMPTY:
		m_tag_explicit_empty = true;
		break;
	case HTMLLexer::TOK_TAG_END:
		eventOpeningTag(m_lexer.getCurTag());
		if (m_tag_explicit_empty) {
			if (m_html_stack.top() && m_cur_tag.getName() == m_html_stack.top()->getName()) {
				eventClosingImportantTag(m_cur_tag.getName());
			} else {
				eventClosingUnimportantTag(m_cur_tag.getName());
			}
		}
		break;
	case HTMLLexer::TOK_CLOSING_TAG_NAME:
		m_closing_tag = m_lexer.tokText();
		break;
	case HTMLLexer::TOK_CLOSING_TAG_BEGIN:
		if (!m_cur_text.empty()) {
			m_html_stack.appendText(m_cur_text);
			m_cur_text.erase();
		}
		break;
	case HTMLLexer::TOK_CLOSING_TAG_END:
		{
			string tname = strtolower(m_closing_tag);
			if (m_html_stack.top() && tname == m_html_stack.top()->getName()) {
				eventClosingImportantTag(tname);
			} else {
				eventClosingUnimportantTag(tname);
				string closing_tag = string("</")+m_closing_tag+'>';
				if (m_html_stack.top()) {
					m_cur_text += closing_tag;
				} else {
					m_processed_data += closing_tag;
				}
			}
		}
		break;
	case HTMLLexer::TOK_ATTR_NAME:
		m_attr_name = strtolower(m_lexer.tokText());
		m_cur_tag.setBooleanAttribute(m_attr_name);
		break;
	case HTMLLexer::TOK_ATTR_VALUE:
		m_cur_tag.setRawAttribute(m_attr_name, m_lexer.tokText().c_str());
		break;
	default:
		break;
	}
}

void
HTMLProcessor::eventOpeningTag(const string& tag_text)
{
	string ttext(tag_text);
	string tname(m_cur_tag.getName());
	bool changed = false;
	if (m_cur_tag.hasAttribute("onmouseover")) {
		string js_output;
		int popup_counter = 0;
		JSOutputCollector js_collector(js_output, popup_counter);
		m_js_env->attachListener(&js_collector);
		m_js_env->executeScriptAsFunction(m_cur_tag.getAttribute("onmouseover"));
		m_js_env->detachListener(&js_collector);
		if (popup_counter > 0) {
			m_cur_tag.setRawAttribute("onmouseover", "/* popup detected by BFilter */");
			changed = true;
		}
	}
	if (m_cur_tag.hasAttribute("onmouseout")) {
		string js_output;
		int popup_counter = 0;
		JSOutputCollector js_collector(js_output, popup_counter);
		m_js_env->attachListener(&js_collector);
		m_js_env->executeScriptAsFunction(m_cur_tag.getAttribute("onmouseout"));
		m_js_env->detachListener(&js_collector);
		if (popup_counter > 0) {
			m_cur_tag.setRawAttribute("onmouseout", "/* popup detected by BFilter */");
			changed = true;
		}
	}
	if (changed) {
		ttext = m_cur_tag.openingToString(m_content_type==CONTENT_XHTML, m_tag_explicit_empty);
	}
	if (tname == "head") {
		m_inside_head = true;
	} else if (tname == "body" || tname == "frameset") {
		m_inside_head = false;
		m_onload = m_cur_tag.getAttribute("onload");
		m_onunload = m_cur_tag.getAttribute("onunload");
		changed = false;
		if (!m_onload.empty()) {
			m_cur_tag.setAttribute("onload", "if (bfilter_onload_ok) { "+m_onload+" }");
			changed = true;
		}
		if (!m_onunload.empty()) {
			m_cur_tag.setAttribute("onunload", "if (bfilter_onunload_ok) { "+m_onunload+" }");
			changed = true;
		}
		if (changed) {
			ttext = m_cur_tag.openingToString(m_content_type==CONTENT_XHTML, m_tag_explicit_empty);
		}
	} else if (tname == "form") {
		string name = m_cur_tag.getAttribute("name");
		if (name.empty())
			name = m_cur_tag.getAttribute("id");
		m_js_env->addForm(name, m_cur_tag.getAttribute("action"), m_cur_tag.getAttribute("target"));
	} else if (tname == "script") {
		URI url(m_cur_tag.getAttribute("src"));
		m_script_3rd_party = isExternalURL(url);
	} else if (m_inside_head && tname == "base") {
		URI url(m_cur_tag.getAttribute("href"));
		if (url.isAbsolute()) {
			m_base_url = url;
		}
	} else if (tname == "img") {
		string imgurl_str = m_cur_tag.getAttribute("src");
		URI imgurl(m_base_url, URI(imgurl_str));
		int img_ad_rule = Rules::getInstance().getAdParam(imgurl.toString());
		bool external_img = isExternalURL(imgurl);
		bool suspicious_img = isSuspiciousURL(imgurl);
		int width = str2num<int>(m_cur_tag.getAttribute("width"), 0);
		int height = str2num<int>(m_cur_tag.getAttribute("height"), 0);
		bool size_defined = (width > 0 && height > 0);
		bool webbug = false;
		if (size_defined && width == 1 && height == 1 && img_ad_rule != 1 && (external_img || suspicious_img)) {
			webbug = true;
			if (m_is_subprocessor)
				throw AdException(AdException::T_WEBBUG);
			URI subst_url(imgurl);
			subst_url.removeFragment();
			subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
			subst_url.setRawQuery("w=1&h=1");
			m_cur_tag.setAttribute("src", subst_url.toString());
			ttext = m_cur_tag.openingToString(m_content_type==CONTENT_XHTML, m_content_type==CONTENT_XHTML);
		}
		HTMLElementNode *tag_a = 0;
		if (!webbug && m_html_stack.top() && (tag_a = m_html_stack.top()->findAncestor("a", true))) {
			if (m_ignored_a_tags.find(tag_a) == m_ignored_a_tags.end()) {
				string aurl_str = tag_a->getAttribute("href");
				URI aurl(m_base_url, URI(aurl_str));
				bool external_link = isExternalURL(aurl);
				bool suspicious_link = isSuspiciousURL(aurl);
				string target = tag_a->getAttribute("target");
				bool is_ad = false;
				bool need_more_info = false;
				if (size_defined) {
					int res = imageIsAd(img_ad_rule, width, height, target, external_link,
					                    external_img, suspicious_link, suspicious_img);
					if (res == 1) {
						is_ad = true;
					} else if (res == 0) {
						need_more_info = true;
					}
				} else { // size not defined
					need_more_info = true;
				}
				if (need_more_info) {
					if (m_is_subprocessor) {
						// TODO: a better solution is needed here
						if (size_defined) {
							throw AdException(AdException::T_IMG, aurl_str, width, height);
						} else {
							int w = size_defined ? width : 311;
							int h = size_defined ? height : 41;
							// 311x41 is not a common ad size, but couldBeAdSize() returns true for it
							if (imageIsAd(img_ad_rule, w, h, target, external_link,
							    external_img, suspicious_link, suspicious_img) != -1) {
								throw AdException(AdException::T_IMG, aurl_str);
							}
						}
					}
					URI subst_url(imgurl);
					subst_url.removeFragment();
					subst_url.setAbsoluteRawPath("bf.intercept-request/image_analyze");
					ostringstream strm;
					strm << "url=" << URI::escape(imgurl.toString()) << "&w=" << width
					<< "&h=" << height << "&iar=" << img_ad_rule << "&ea="
					<< (external_link?1:0) << "&ei=" << (external_img?1:0)
					<< "&sa=" << (suspicious_link?1:0) << "&si=" << (suspicious_img?1:0)
					<< "&target=" << URI::escape(target);
					subst_url.setRawQuery(strm.str());
					m_cur_tag.setAttribute("src", subst_url.toString());
					ttext = m_cur_tag.openingToString(m_content_type==CONTENT_XHTML, m_content_type==CONTENT_XHTML);
				} else if (is_ad) {
					if (m_is_subprocessor)
						throw AdException(AdException::T_IMG, aurl_str, width, height);
					URI subst_url(imgurl);
					subst_url.removeFragment();
					subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
					subst_url.setRawQuery("w="+num2str(width)+"&h="+num2str(height));
					m_cur_tag.setAttribute("src", subst_url.toString());
					ttext = m_cur_tag.openingToString(m_content_type==CONTENT_XHTML, m_content_type==CONTENT_XHTML);
					HTMLElementNode *tag_iframe = tag_a->findAncestor("iframe");
					if (tag_iframe) {
						tag_iframe->setBooleanAttribute("ad");
						// will be processed when the iframe closes
					}
				}
			}
		}
	} else if (tname == "table" || tname == "tr" || tname == "td"
	           || tname == "div" || tname == "p" || tname == "center") {
		// These tags can't be inside an A tag.
		// If they are, then the A tag is considered unclosed and is ignored by the filtering code.
		if (m_html_stack.top()) {
			const HTMLElementNode *tag_a = m_html_stack.top()->findAncestor("a", true);
			if (tag_a) {
				m_ignored_a_tags.insert(tag_a);
			}
		}
	}
	
	if (m_important_tags.find(tname) != m_important_tags.end()) {
		if (HTMLTagInfo::getInfoOn(tname).m_etag_status != HTMLTagInfo::ETAG_FORBIDDEN) {
			m_html_stack.pushElement(m_cur_tag);
		} else {
			if (m_html_stack.top()) {
				m_html_stack.appendElement(m_cur_tag);
			} else {
				m_processed_data += ttext;
			}
		}
	} else {
		if (m_html_stack.top()) {
			m_cur_text += ttext;
		} else {
			m_processed_data += ttext;
		}
	}
}

void
HTMLProcessor::eventClosingUnimportantTag(const string& tname)
{
	if (tname == "head") {
		m_inside_head = false;
	} else if (tname == "body" || tname == "frameset") {
		if (!m_onload_check_done) {
			bool onload_ok = true;
			bool onunload_ok = true;
			if (!m_onload.empty()) {
				string js_output;
				int popup_counter = 0;
				JSOutputCollector js_collector(js_output, popup_counter);
				m_js_env->attachListener(&js_collector);
				m_js_env->executeScriptAsFunction(m_onload);
				m_js_env->detachListener(&js_collector);
				onload_ok = (popup_counter == 0);
			}
			if (!m_onunload.empty()) {
				string js_output;
				int popup_counter = 0;
				JSOutputCollector js_collector(js_output, popup_counter);
				m_js_env->attachListener(&js_collector);
				m_js_env->executeScriptAsFunction(m_onunload);
				m_js_env->detachListener(&js_collector);
				onunload_ok = (popup_counter == 0);
			}
			if (!m_onload.empty() || !m_onunload.empty()) {
				string data = "<script type=\"text/javascript\"> ";
				data += string("bfilter_onload_ok = ")+(onload_ok ? "true" : "false")+"; ";
				data += string("bfilter_onunload_ok = ")+(onunload_ok ? "true" : "false")+"; ";
				data += string("</script>\r\n");
				if (m_html_stack.top()) {
					m_cur_text += data;
				} else {
					m_processed_data += data;
				}
			}
			m_onload_check_done = true;
		}
	}
}

// this function closes the top tag in m_html_stack
void
HTMLProcessor::eventClosingImportantTag(const string& tname)
{
	bool tag_removed = false;
	if (tname == "a") {
		if (m_target_blank == 1 && m_html_stack.top()->getAttribute("target") == "_blank") {
			m_html_stack.top()->removeAttribute("target");
		}
	} else if (tname == "iframe") {
		string url_str = m_html_stack.top()->getAttribute("src");
		URI url(m_base_url, URI(url_str));
		int width = str2num<int>(m_html_stack.top()->getAttribute("width"));
		int height = str2num<int>(m_html_stack.top()->getAttribute("height"));
		int ad_rule = Rules::getInstance().getAdParam(url.toString());
		bool external_url = isExternalURL(url);
		bool is_ad = external_url || m_html_stack.top()->hasAttribute("ad")
			|| (isSuspiciousURL(url) && isCommonAdSize(width, height));
		if ((is_ad || ad_rule==2) && ad_rule!=1) {
			
			if (m_is_subprocessor)
				throw AdException(AdException::T_IFRAME, url_str, width, height);
			m_html_stack.popElement();
			tag_removed = true;
			HTMLStack rstack;
			HTMLElementNode tag_a("a");
			tag_a.setAttribute("href", url_str);
			tag_a.setRawAttribute("alt", "A replacement for an iframe ad");
			HTMLElementNode tag_img("img");
			URI subst_url(url);
			subst_url.removeFragment();
			subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
			subst_url.setRawQuery("w="+num2str(width)+"&h="+num2str(height));
			tag_img.setRawAttribute("src", subst_url.toString());
			tag_img.setRawAttribute("border", "0");
			if (width > 0 && height > 0) {
				tag_img.setRawAttribute("width", num2str(width));
				tag_img.setRawAttribute("height", num2str(height));
			}
			rstack.pushElement(tag_a);
			rstack.pushElement(tag_img);
			string replace_str = rstack.bottom()->toString(m_content_type==CONTENT_XHTML);
			if (m_html_stack.top()) {
				m_html_stack.appendText(replace_str);
			} else {
				m_processed_data += replace_str;
			}
		} else if (is_ad) {
			m_html_stack.top()->removeAttribute("ad");
		}
	} else if (tname == "object") {
		if (strtolower(m_html_stack.top()->getAttribute("classid")) == "clsid:d27cdb6e-ae6d-11cf-96b8-444553540000") {
			list<HTMLNode *> children = m_html_stack.top()->getChildren();
			list<HTMLNode *>::iterator iter = children.begin();
			for (; iter != children.end(); ++iter) {
				if ((*iter)->getType() != HTMLNode::NODE_ELEMENT) {
					continue;
				}
				HTMLElementNode *node = dynamic_cast<HTMLElementNode *>(*iter);
				if (node->getName() != "param") {
					continue;
				}
				if (strtolower(node->getAttribute("name")) != "movie") {
					continue;
				}
				int width = str2num<int>(m_html_stack.top()->getAttribute("width"));
				int height = str2num<int>(m_html_stack.top()->getAttribute("height"));
				string url_str = node->getAttribute("value");
				URI url(m_base_url, URI(url_str));
				int ad_rule = Rules::getInstance().getAdParam(url.toString());
				bool external = isExternalURL(url);
				bool is_ad = external;
				if (!is_ad && isSuspiciousURL(url)) {
					if (isCommonAdSize(width, height) || couldBeAdSize(width, height)) {
						is_ad = true;
					}
				}
				if ((is_ad || ad_rule==2) && ad_rule!=1) {
					if (m_is_subprocessor)
						throw AdException(AdException::T_FLASH, url_str, width, height);
					m_html_stack.popElement();
					tag_removed = true;
					string replace_str;
					if (width > 2 && height > 2) {
						HTMLStack rstack;
						HTMLElementNode tag_a("a");
						tag_a.setAttribute("href", url_str);
						tag_a.setRawAttribute("alt", "A replacement for a flash ad");
						HTMLElementNode tag_img("img");
						URI subst_url(url);
						subst_url.removeFragment();
						subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
						subst_url.setRawQuery("w="+num2str(width)+"&h="+num2str(height));
						tag_img.setAttribute("src", subst_url.toString());
						tag_img.setRawAttribute("border", "0");
						tag_img.setRawAttribute("width", num2str(width));
						tag_img.setRawAttribute("height", num2str(height));
						rstack.pushElement(tag_a);
						rstack.pushElement(tag_img);
						replace_str = string("<!--\r\nFlash ad removed by BFilter [replacement follows]\r\n-->")+
							rstack.bottom()->toString(m_content_type==CONTENT_XHTML);
					} else {
						replace_str = "<!--\r\nFlash ad removed by BFilter\r\n-->";
					}
					if (m_html_stack.top()) {
						m_html_stack.appendText(replace_str);
					} else {
						m_processed_data += replace_str;
					}
					tag_removed = true;
				}
				break;
			}
		}
	} else if (tname == "script") {
		try {
			if (!State::getInstance().getScriptFiltering()) {
				throw true;
			}
			string script_code;
			string script_src;
			int ad_rule = 0;
			bool script_is_js = (!m_html_stack.top()->hasAttribute("type") ||
				strtolower(m_html_stack.top()->getAttribute("type")) == "text/javascript") &&
				(!m_html_stack.top()->hasAttribute("language") ||
				strtolower(m_html_stack.top()->getAttribute("language")).find("javascript") == 0);
			bool script_external = m_html_stack.top()->hasAttribute("src");
			if (!script_external) {
				DEBUGLOG("processing inline script");
				script_code = m_html_stack.top()->contentsToString(m_content_type==CONTENT_XHTML);
			} else {
				script_src = m_html_stack.top()->getAttribute("src");
				DEBUGLOG("Fetching the external script from " << script_src);
				HttpFetcher fetcher;
				URI url(m_base_url, URI(script_src));
				ad_rule = Rules::getInstance().getAdParam(url.toString());
				if (ad_rule == 2) {
					if (m_is_subprocessor) {
						throw AdException(AdException::T_SCRIPT_RULES);
					} else {
						throw string("<!--\r\nScript removed by BFilter (ad rule)\r\n-->");
					}
				}
				try {
					for (int i = 0;; i++) {
						if (i >= MAX_SCRIPT_FETCH_REDIRECTS) {
							throw HttpFetcher::Exception("too many redirects ("+num2str(i)+")");
						}
						HttpRequest request("GET", url);
						request.headers().setHeader(HttpHeader("Referrer", m_base_url.toString()));
						request.headers().setHeader(HttpHeader("Accept", "*" "/" "*"));
						HttpResponse response = fetcher.request(request);
						int rcode = response.getStatusCode();
						if ((rcode >= 301 && rcode <= 303) || rcode == 307) {
							url = URI(m_base_url, URI(response.headers().getHeader("Location").getValue()));
							continue;
						} else if (rcode != 200) {
							throw HttpFetcher::Exception(num2str(rcode)+" "+response.getStatusMessage());
						} else {
							script_code = response.body();
							break;
						}
					}
				} catch (HttpFetcher::Exception& e) {
					string text = "<!--\r\nBFilter: Error fetching the script from ";
					text += url.toString()+"\r\nReason: ";
					text += e.getMessage()+"\r\n-->";
					DEBUGLOG("Fetching the script failed: " << e.getMessage());
					throw text;
				}
				DEBUGLOG("External script fetched successfully");
			}
			if (script_is_js && m_jsengine == 1 && m_nest_level < MAX_NEST_LEVEL && ad_rule != 1) {
				string js_version;
				string javascript("javascript");
				string::size_type js_pos = m_html_stack.top()->getAttribute("language").find(javascript);
				if (js_pos != string::npos) {
					js_version = m_html_stack.top()->getAttribute("language");
					js_version.erase(0, js_pos+javascript.length());
				}
				
				string js_output;
				int popup_counter = 0;
				JSOutputCollector js_collector(js_output, popup_counter);
				m_js_env->attachListener(&js_collector);
				m_js_env->executeScript(script_code, js_version);
				m_js_env->detachListener(&js_collector);
				
				if (m_webbug_open_comment &&
				    (js_output == "-->" || js_output == "-->\n" || js_output == "-->\r\n")) {
					// don't throw AdException - this situation should not happen in subscripts,
					// but if it does, we better don't interfere
					m_webbug_open_comment = false;
					throw string("<!--\r\nScript removed by BFilter (jsengine -> webbug tail)\r\n-->");
				}
				
				if (popup_counter > 0) {
					DEBUGLOG("The script opens a popup");
					if (js_output.rfind("<!--") >= js_output.length()-7) {
						m_webbug_open_comment = true;
					}
					if (m_is_subprocessor) {
						throw AdException(AdException::T_POPUP);
					} else {
						throw string("<!--\r\nScript removed by BFilter (jsengine -> popup)\r\n-->");
					}
				}
				
				DEBUGLOG("Launching subprocessor to examine the output of the script (nest_level=" << m_nest_level+1 << ")");
				Context context;
				saveContext(context);
				HTMLProcessor child_proc(m_owner, context, m_js_env, m_nest_level+1);
				
				try {
					child_proc.consumeData(js_output);
					child_proc.consumeEOF();
					DEBUGLOG("Subprocessor finishes - no ads found");
				} catch (AdException& e) {
					DEBUGLOG("Subprocessor has caught something: " << e.getInfo());
					if (js_output.rfind("<!--") >= js_output.length()-7) {
						m_webbug_open_comment = true;
					}
					if (m_is_subprocessor) {
						throw AdException(AdException::T_JSENGINE, e);
					} else {
						int width = e.getWidth();
						int height = e.getHeight();
						bool do_replacement = (!e.getURL().empty() &&
							(width > 2 || width == 0) && (height > 2 || height == 0));
						string replace_str = string("<!--\r\nScript removed by BFilter (jsengine -> ")+
						e.getInfo()+")"+(do_replacement ? "[replacement follows]" : "")+" \r\n-->";
						if (do_replacement) {
							if (width == 0)
								width = 10;
							if (height == 0)
								height = 10;
							tag_removed = true;
							HTMLStack rstack;
							HTMLElementNode tag_a("a");
							tag_a.setAttribute("href", e.getURL());
							tag_a.setRawAttribute("alt", "A replacement for a js ad");
							HTMLElementNode tag_img("img");
							URI subst_url(e.getURL());
							subst_url.removeFragment();
							subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
							subst_url.setRawQuery("w="+num2str(width)+"&h="+num2str(height));
							tag_img.setAttribute("src", subst_url.toString());
							tag_img.setRawAttribute("border", "0");
							tag_img.setRawAttribute("width", num2str(width));
							tag_img.setRawAttribute("height", num2str(height));
							rstack.pushElement(tag_a);
							rstack.pushElement(tag_img);
							string rhtml = rstack.bottom()->toString(m_content_type==CONTENT_XHTML);
							replace_str += string("<script type=\"text/javascript\">\r\n<!--\r\n")+
								"document.writeln(unescape('"+URI::escape(rhtml)+"'));\r\n//--></script>";
						}
						throw replace_str;
					}
				}
			} else {
				if (ad_rule != 1 && !allowScript(script_code)) {
					if (m_is_subprocessor) {
						throw AdException(AdException::T_SCRIPT_RULES);
					} else {
						throw string("<!--\r\nScript removed by BFilter (script rule)\r\n-->");
					}
				}
			}
			m_webbug_open_comment = false;
			
			// no string exceptions after this point!
			if (script_external) {
				bool safe = (strtolower(script_code).find("</script") == string::npos);
				bool leave_as_is = false;
				string script_body;
				if (safe) {
					script_body = script_code;
				} else if (script_is_js) {
					script_body = string("\r\neval(unescape('")+
						URI::escape(script_code)+"'))\r\n;";
				} else {
					leave_as_is = true;
				}
				if (!leave_as_is) {
					m_html_stack.top()->removeAttribute("src");
					m_html_stack.clearTop();
					m_html_stack.appendText(script_body);
					string content = m_html_stack.top()->toString(m_content_type==CONTENT_XHTML);
					string replacement = string("<!-- The following script has been fetched from ")+
						str_replace(">", "&gt;", script_src)+
						(safe ? "" : " and escaped")+" by BFilter -->"+content;
					m_html_stack.popElement();
					if (m_html_stack.top()) {
						m_html_stack.appendText(replacement);
					} else {
						m_processed_data += replacement;
					}
					tag_removed = true;
				}
			}
		} catch (string& e) {
			m_html_stack.popElement();
			if (m_html_stack.top()) {
				m_html_stack.appendText(e);
			} else {
				m_processed_data += e;
			}
			tag_removed = true;
		} catch (bool e) {}
	}
	
	if (!tag_removed) {
		if (m_html_stack.top() == m_html_stack.bottom()) {
			m_processed_data += m_html_stack.top()->toString(m_content_type==CONTENT_XHTML);
			m_html_stack.popElement();
		} else {
			m_html_stack.walkOn();
		}
	}
}

bool
HTMLProcessor::allowScript(const string& script)
{
	switch (m_scripts) {
	case 0:
		return true;
	case 1:
		return (m_inside_head || !m_script_3rd_party);
	case 2:
		return !m_script_3rd_party;
	case 3:
		return (m_inside_head || (!m_script_3rd_party && script.find(".write")==string::npos));
	case 4:
		return (m_inside_head || (!m_script_3rd_party && script.find("function ")!=string::npos));
	case 5:
		return m_inside_head;
	case 6:
		return (m_inside_head && !m_script_3rd_party);
	case 7:
		return false;
	default:
		return true;
	}
}

void
HTMLProcessor::constructImportantTags()
{
	// Important tags are tags in which we may modify their contents or remove them altogether.
	// These tags are placed to m_html_stack.
	// The tags listed here MUST be present in class HTMLTagInfo
	m_important_tags.insert("a");
	m_important_tags.insert("iframe");
	m_important_tags.insert("script");
	m_important_tags.insert("img");
	m_important_tags.insert("object");
	m_important_tags.insert("param");
}

