/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef HTMLPROCESSOR_H_
#define HTMLPROCESSOR_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <set>

#include <htmllexer.h>
#include "htmlstack.h"
#include "uri.h"
#include "jsenvironment.h"
#include "jsenvproxy.h"
#include "debug.h"

using namespace std;

class BFilter;

class HTMLProcessor
{
public:
	enum content_type_enum { CONTENT_HTML, CONTENT_XHTML };
	class Context
	{
		friend class HTMLProcessor;
	public:
		Context() : m_base_url("http://dummy.url") {}
	private:
		content_type_enum m_content_type;
		URI m_base_url;
		bool m_inside_head;
		bool m_script_3rd_party;
		int m_scripts;
		int m_jsengine;
		int m_target_blank;
		HTMLStack m_html_stack;
		string m_onload;
		string m_onunload;
	};
	class AdException
	{
	public:
		enum type { T_IMG, T_IFRAME, T_FLASH, T_WEBBUG, T_POPUP, T_JSENGINE, T_SCRIPT_RULES };
		AdException(type t, string url="", int width=0, int height=0)
		: m_type(t), m_info(typeToInfo(t)), m_url(url), m_width(width), m_height(height) {}
		AdException(type t, const AdException& child)
		: m_type(t), m_info(typeToInfo(t)+" -> "+child.getInfo()),
		  m_url(child.m_url), m_width(child.m_width), m_height(child.m_height) {}
		type getType() const { return m_type; }
		const string& getInfo() const { return m_info; }
		const string& getURL() const { return m_url; }
		int getWidth() const { return m_width; }
		int getHeight() const { return m_height; }
	private:
		string typeToInfo(type t)
		{
			switch (t) {
			case T_IMG:
				return "img ad";
			case T_IFRAME:
				return "iframe ad";
			case T_FLASH:
				return "flash ad";
			case T_WEBBUG:
				return "webbug";
			case T_POPUP:
				return "popup";
			case T_JSENGINE:
				return "jsengine";
			case T_SCRIPT_RULES:
				return "script rules";
			default:
				return "unknown";
			}
		}
		type m_type;
		string m_info;
		string m_url;
		int m_width;
		int m_height;
	};
	HTMLProcessor(BFilter *owner);
	HTMLProcessor(BFilter *owner, const Context& context, JSEnvProxy& js_env, int nest_level);
	void setBaseURL(const URI& url);
	void consumeData(const string& data);
	void consumeEOF();
	void reset();
	const string& peekProcessedData() const { return m_processed_data; }
	string readProcessedData()
		{ string tmp = m_processed_data; m_processed_data.erase(); return tmp; }
	void saveContext(Context& context) const;
	static bool isCommonAdSize(int width, int height);
	static bool couldBeAdSize(int width, int height);
	static int imageIsAd(int ad_rule, int width, int height, const string& target,
		bool external_link, bool external_img, bool suspicious_link, bool suspicious_img,
		int animated=0, int redir=0, int cookie=0);
private:
	static const int MAX_SCRIPT_FETCH_REDIRECTS = 2;
	static const int MAX_NEST_LEVEL = 2;
	class JSOutputCollector : public JSListener
	{
	public:
		JSOutputCollector(string& buffer, int& popup_counter)
		: m_buffer(buffer), m_popup_counter(popup_counter) {}
		virtual void processDataChunk(const string& data) {
			m_buffer += data;
		}
		virtual void processPopup()
		{
			++m_popup_counter;
		}
	private:
		string& m_buffer;
		int& m_popup_counter;
	};
	bool isExternalURL(const URI& url);
	bool isSuspiciousURL(const URI& url);
	void processToken(int tok);
	void eventOpeningTag(const string& tag_text);
	void eventClosingUnimportantTag(const string& tname);
	void eventClosingImportantTag(const string& tname);
	bool allowScript(const string& script);
	void constructImportantTags();
	bool m_is_subprocessor;
	BFilter *m_owner;
	content_type_enum m_content_type;
	HTMLLexer m_lexer;
	string m_processed_data;
	URI m_base_url;
	string m_closing_tag;
	string m_attr_name;
	HTMLElementNode m_cur_tag;
	string m_cur_text;
	bool m_tag_explicit_empty;
	bool m_inside_head;
	bool m_script_3rd_party;
	bool m_onload_check_done;
	int m_scripts; // rule value from config
	int m_jsengine; // jsengine value from config
	int m_target_blank; // target_blank value from config
	HTMLStack m_html_stack;
	set<const HTMLElementNode *> m_ignored_a_tags; // used for handling of unclosed tags
	string m_onload;
	string m_onunload;
	JSEnvProxy m_js_env;
	int m_nest_level;
	bool m_webbug_open_comment; // true if the last script was rejected, and its output ends with <!--
	set<string> m_important_tags;
};

#endif
