/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstatefooters.h"

#include <string>
#include "httpstream.h"
#include "httpheader.h"
#include "httpheaderscollection.h"
#include "utils.h"

using namespace std;

HttpStateFooters::HttpStateFooters(HttpStream& owner)
:	m_owner(owner), m_waiting_lws(false)
{}

void
HttpStateFooters::consumeDataChunk(string& pool)
{
	const string* pdata = &pool;
	string accumulated_data;
	string::size_type queued_data_len = m_queued_data.length();
	if (queued_data_len) {
		accumulated_data = m_queued_data+pool;
		pdata = &accumulated_data;
		m_queued_data.erase();
	}
	string::size_type pos = 0, pos2;
	while ((pos2 = pdata->find("\r\n", pos)) != string::npos) {
		if (feedLine(pdata->substr(pos, pos2-pos))) {
			pos2 += 2;
			if (pos2 > queued_data_len) {
				pool.erase(0, pos2-queued_data_len);
			}
			m_owner.switchState(HttpStream::STATE_INITIAL);
			m_owner.eventMessageEnds(m_owner.m_connection_needs_closing);
			return;
		} else {
			pos2 += 2;
		}
		pos = pos2;
	}
	m_queued_data.append(*pdata, pos, string::npos);
	pool.assign("");
}

void
HttpStateFooters::consumeEOF()
{
	m_owner.eventError(HttpStream::ERR_UNEXPECTED_EOF);
	m_owner.jumpToState(HttpStream::STATE_ERROR);
}

void
HttpStateFooters::cleanup()
{
	m_headers.clear();
	m_cur_header.erase();
	m_queued_data.erase();
	m_waiting_lws = false;
}

bool
HttpStateFooters::feedLine(const string& line)
{
	if (line.empty()) {
		if (m_waiting_lws)
			addHeader();
		return true;
	}
	
	if (m_waiting_lws) {
		int lws_count = 0;
		bool lws_end = false;
		string::const_iterator iter = line.begin();
		for (; iter != line.end(); ++iter, ++lws_count) {
			const char ch = *iter;
			if (ch != ' ' && ch != '\t') {
				lws_end = true;
				break;
			}
		}
		if (!lws_end) {
			return false;
		} else {
			if (lws_count > 0) {
				m_cur_header += ' '+line.substr(lws_count);
			} else {
				addHeader();
				m_cur_header += line;
			}
		}
	} else {
		m_cur_header += line;
		m_waiting_lws = true;
	}
	return false;
}

void
HttpStateFooters::addHeader()
{
	HttpHeader hdr(m_cur_header);
	string name = hdr.getName();
	if (name.empty()) {
		m_owner.eventError(HttpStream::ERR_PARSING_HEADER);
	} else {
		HttpHeader *hptr = m_headers.getHeaderPtr(name);
		if (hptr) {
			hptr->appendValue(hdr.getValue());
		} else {
			m_headers.setHeader(hdr);
		}
	}
	m_cur_header.erase();
}

