/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstateheaders.h"

#include <string>
#include <cstdlib>
#include "httpstream.h"
#include "httpheader.h"
#include "httpheaderscollection.h"
#include "httpversion.h"
#include "httprequestmetadata.h"
#include "httpresponsemetadata.h"
#include "httprequestline.h"
#include "httpstatusline.h"
#include "utils.h"
#include "debug.h"

using namespace std;

HttpStateHeaders::HttpStateHeaders(HttpStream& owner)
:	m_owner(owner), m_waiting_lws(false), m_request_line(0), m_status_line(0)
{}

HttpStateHeaders::~HttpStateHeaders()
{
	if (m_request_line) {
		delete m_request_line;
	}
	if (m_status_line)
		delete m_status_line;
}

void
HttpStateHeaders::consumeDataChunk(string& pool)
{
	const string* pdata = &pool;
	string accumulated_data;
	string::size_type queued_data_len = m_queued_data.length();
	if (queued_data_len) {
		accumulated_data = m_queued_data+pool;
		pdata = &accumulated_data;
		m_queued_data.erase();
	}
	string::size_type pos = 0, pos2;
	while ((pos2 = pdata->find("\r\n", pos)) != string::npos) {
		if (feedLine(pdata->substr(pos, pos2-pos))) {
			pos2 += 2;
			if (pos2 > queued_data_len) {
				pool.erase(0, pos2-queued_data_len);
			}
			headersReceived();
			return;
		} else {
			pos2 += 2;
		}
		pos = pos2;
	}
	m_queued_data.append(*pdata, pos, string::npos);
	pool.assign("");
}

void
HttpStateHeaders::consumeEOF()
{
	m_owner.eventError(HttpStream::ERR_UNEXPECTED_EOF);
	m_owner.jumpToState(HttpStream::STATE_ERROR);
}

void
HttpStateHeaders::cleanup()
{
	m_headers.clear();
	m_cur_header.erase();
	m_queued_data.erase();
	m_waiting_lws = false;
	if (m_request_line) {
		delete m_request_line;
		m_request_line = 0;
	}
	if (m_status_line) {
		delete m_status_line;
		m_status_line = 0;
	}
}

void
HttpStateHeaders::setRequestLine(const HttpRequestLine& request_line)
{
	if (m_request_line) {
		delete m_request_line;
	}
	m_request_line = new HttpRequestLine(request_line);
}

void
HttpStateHeaders::setStatusLine(const HttpStatusLine& status_line)
{
	if (m_status_line) {
		delete m_status_line;
	}
	m_status_line = new HttpStatusLine(status_line);
}

bool
HttpStateHeaders::feedLine(const string& line)
{
	if (line.empty()) {
		if (m_waiting_lws)
			addHeader();
		return true;
	}
	
	if (m_waiting_lws) {
		int lws_count = 0;
		bool lws_end = false;
		string::const_iterator iter = line.begin();
		for (; iter != line.end(); ++iter, ++lws_count) {
			const char ch = *iter;
			if (ch != ' ' && ch != '\t') {
				lws_end = true;
				break;
			}
		}
		if (!lws_end) {
			return false;
		} else {
			if (lws_count > 0) {
				m_cur_header += ' '+line.substr(lws_count);
			} else {
				addHeader();
				m_cur_header += line;
			}
		}
	} else {
		m_cur_header += line;
		m_waiting_lws = true;
	}
	return false;
}

void
HttpStateHeaders::addHeader()
{
	HttpHeader hdr(m_cur_header);
	string name = hdr.getName();
	if (name.empty()) {
		m_owner.eventError(HttpStream::ERR_PARSING_HEADER);
	} else {
		HttpHeader *hptr = m_headers.getHeaderPtr(name);
		if (hptr) {
			hptr->appendValue(hdr.getValue());
		} else {
			m_headers.setHeader(hdr);
		}
	}
	m_cur_header.erase();
}

void
HttpStateHeaders::headersReceived()
{
	bool conn_close = (m_headers.getHeader("Connection").elements().hasElement("close")
		|| m_headers.getHeader("Proxy-Connection").elements().hasElement("close"));
	bool keep_alive = !conn_close && ((m_headers.getHeader("Connection").elements().hasElement("keep-alive")
		|| m_headers.getHeader("Proxy-Connection").elements().hasElement("keep-alive")));
	
	if (m_owner.isRequestStream()) {
		if (!m_request_line) {
			DEBUGLOG("BUG");
			abort();
		} else {
			if (m_request_line->getHttpVersion() < HttpVersion::HTTP_1_1 && !keep_alive) {
				conn_close = true;
			}
		}
	} else {
		if (!m_status_line) {
			DEBUGLOG("BUG");
			abort();
		} else {
			if (m_status_line->getHttpVersion() < HttpVersion::HTTP_1_1 && !keep_alive) {
				conn_close = true;
			}
		}
	}
	
	
	if (!m_owner.isRequestStream() && (m_owner.isResponseOnHeadRequest() ||
	    (m_status_line->getCode() >= 100 && m_status_line->getCode() < 200))) {
		HttpResponseMetadata metadata(*m_status_line, m_headers);
		metadata.setBodyStatus(HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
		m_owner.eventGotMetadata(metadata);
		m_owner.switchState(HttpStream::STATE_INITIAL);
		m_owner.eventMessageEnds(conn_close);
		return;
	}
	
	bool must_not_have_body = false;
	
	if (m_owner.isRequestStream()) {
		string method = m_request_line->getMethod();
		if (method == "GET" || method == "HEAD") {
			must_not_have_body = true;
		}
	} else {
		int code = m_status_line->getCode();
		if (code == 304 || (code >= 100 && code < 200)) {
			must_not_have_body = true;
		}
	}
	
	if (must_not_have_body) {
		if (m_owner.isRequestStream()) {
			HttpRequestMetadata metadata(*m_request_line, m_headers);
			metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
			metadata.setBodySize(0);
			m_owner.eventGotMetadata(metadata);
			m_owner.switchState(HttpStream::STATE_INITIAL);
			m_owner.eventMessageEnds(false);
		} else {
			HttpResponseMetadata metadata(*m_status_line, m_headers);
			metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
			metadata.setBodySize(0);
			m_owner.eventGotMetadata(metadata);
			m_owner.switchState(HttpStream::STATE_INITIAL);
			m_owner.eventMessageEnds(conn_close);
		}
		return;
	}
	
	HttpMessageMetadata::BodyStatus body_status = HttpMessageMetadata::BODY_SIZE_UNKNOWN;
	string::size_type body_size = 0;
	
	if (m_headers.hasHeader("Content-Length")) {
		body_status = HttpMessageMetadata::BODY_SIZE_KNOWN;
		body_size = str2num<string::size_type>(m_headers.getHeader("Content-Length").getValue());
	}
	
	bool chunked_body = false;
	HttpHeader *tenc_header = m_headers.getHeaderPtr("Transfer-Encoding");
	if (tenc_header) {
		if (tenc_header->elements().empty()) {
			m_headers.removeHeader("Transfer-Encoding");
		} else {
			string last_encoding = tenc_header->elements().getLastElement()->getName();
			chunked_body = (last_encoding == "chunked");
			if (chunked_body || last_encoding == "identity") {
				tenc_header->elements().removeLastElement();
				if (tenc_header->elements().empty()) {
					m_headers.removeHeader("Transfer-Encoding");
				}
			}
		}
	}
	
	if (chunked_body) {
		body_status = HttpMessageMetadata::BODY_SIZE_UNKNOWN;
		m_headers.removeHeader("Content-Length");
	}
	
	if (!chunked_body && body_status == HttpMessageMetadata::BODY_SIZE_UNKNOWN) {
		if (m_owner.isRequestStream()) {
			if (m_request_line->getMethod() == "POST" || m_request_line->getMethod() == "PUT") {
				m_owner.eventError(HttpStream::ERR_LENGTH_REQUIRED);
				m_owner.jumpToState(HttpStream::STATE_ERROR);
			} else {
				body_status = HttpMessageMetadata::BODY_SIZE_KNOWN;
				body_size = 0;
			}
		}
	}
	
	m_owner.m_connection_needs_closing = (!m_owner.isRequestStream() && (conn_close
		|| (body_status == HttpMessageMetadata::BODY_SIZE_UNKNOWN && !chunked_body)));
	
	if (m_owner.isRequestStream()) {
		HttpRequestMetadata metadata(*m_request_line, m_headers);
		metadata.setBodyStatus(body_status);
		metadata.setBodySize(body_size);
		m_owner.eventGotMetadata(metadata);
	} else {
		HttpResponseMetadata metadata(*m_status_line, m_headers);
		metadata.setBodyStatus(body_status);
		metadata.setBodySize(body_size);
		m_owner.eventGotMetadata(metadata);
	}
	
	if (chunked_body) {
		m_owner.switchState(HttpStream::STATE_CHUNK_HEADER);
	} else {
		if (body_status == HttpMessageMetadata::BODY_SIZE_KNOWN) {
			m_owner.m_state_flat_body.setBodySize(body_size);
			if (body_size == 0) {
				m_owner.switchState(HttpStream::STATE_INITIAL);
				m_owner.eventMessageEnds(m_owner.isRequestStream() ? false : conn_close);
				return;
			}
		}
		m_owner.switchState(HttpStream::STATE_FLAT_BODY);
	}
}

