/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstatestartline.h"

#include <string>
#include "httpstream.h"
#include "stringtokenizer.h"
#include "uri.h"
#include "utils.h"

using namespace std;

void
HttpStateStartLine::consumeDataChunk(string& pool)
{
	string::size_type pos = pool.find("\r\n");
	if (pos == string::npos) {
		m_consumed_data += pool;
		m_start_line += pool;
		pool.assign("");
	} else {
		m_consumed_data.append(pool, 0, pos+2);
		m_start_line.append(pool, 0, pos);
		pool.erase(0, pos+2);
		if (m_start_line.empty()) {
			m_owner.switchState(HttpStream::STATE_INITIAL);
		} else {
			processStartLine(pool);
		}
	}
}

void
HttpStateStartLine::consumeEOF()
{
	if (m_owner.isRequestStream()) {
		m_owner.eventError(HttpStream::ERR_UNEXPECTED_EOF);
		m_owner.jumpToState(HttpStream::STATE_ERROR);
	} else {
		HttpResponseMetadata metadata(HttpStatusLine(HttpVersion::HTTP_0_9, 200), HttpHeadersCollection());
		metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
		m_owner.eventGotMetadata(metadata);
		m_owner.eventGotContentPart(m_consumed_data);
		m_owner.switchState(HttpStream::STATE_INITIAL);
		m_owner.eventMessageEnds(true);
	}
}

void
HttpStateStartLine::cleanup()
{
	m_consumed_data.erase();
	m_start_line.erase();
}

void
HttpStateStartLine::processStartLine(string& pool)
{
	if (m_owner.isRequestStream()) {
		string::size_type method_end_pos = m_start_line.find_first_of(" \t");
		if (method_end_pos == string::npos) {
			m_owner.eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
			m_owner.jumpToState(HttpStream::STATE_ERROR);
		}
		string method(m_start_line, 0, method_end_pos);
		string::size_type uri_start_pos = m_start_line.find_first_not_of(" \t", method_end_pos+1);
		if (uri_start_pos == string::npos) {
			m_owner.eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
			m_owner.jumpToState(HttpStream::STATE_ERROR);
		}
		string::size_type protocol_before_start_pos = m_start_line.find_last_of(" \t");
		if (protocol_before_start_pos == string::npos) {
			m_owner.eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
			m_owner.jumpToState(HttpStream::STATE_ERROR);
		}
		string protocol(m_start_line, protocol_before_start_pos+1);
		string uri_str = rtrim(string(m_start_line, uri_start_pos, protocol_before_start_pos-uri_start_pos));
		string::size_type pos = 0;
		if (protocol.substr(0, 5) != "HTTP/") {
			m_owner.eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
			m_owner.jumpToState(HttpStream::STATE_ERROR);
		}
		StringTokenizer stk(protocol.substr(5), ".");
		int http_major = str2num<int>(stk.nextToken(), -1);
		int http_minor = str2num<int>(stk.nextToken(), -1);
		if (http_major < 0 || http_minor < 0) {
			m_owner.eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
			m_owner.jumpToState(HttpStream::STATE_ERROR);
		}
		m_owner.m_state_headers.setRequestLine(HttpRequestLine(method, URI(uri_str), HttpVersion(http_major, http_minor)));
		m_owner.switchState(HttpStream::STATE_HEADERS);
	} else {
		string protocol;
		string::size_type pos = 0;
		int http_major = 0;
		int http_minor = 9;
		if (m_start_line.substr(0, 5) == "HTTP/") {
			string::size_type pos1 = m_start_line.find(' ', 5);
			if (pos1 == string::npos) {
				switchToSimpleResponse(pool);
				return;
			}
			protocol.assign(m_start_line, 0, pos1);
			StringTokenizer stk(protocol.substr(5), ".");
			http_major = str2num<int>(stk.nextToken(), -1);
			http_minor = str2num<int>(stk.nextToken(), -1);
			if (http_major < 0 || http_minor < 0) {
				switchToSimpleResponse(pool);
				return;
			}
			pos = pos1+1;
		}
		int code = -1;
		istringstream strm(m_start_line.substr(pos));
		strm >> code;
		if (code < 0) {
			switchToSimpleResponse(pool);
			return;
		}
		string message;
		char buf[100];
		strm.get(); // skip the first space
		while (true) {
			strm.read(buf, sizeof(buf));
			if (strm.gcount() == 0)
				break;
			message.append(buf, buf+strm.gcount());
		}
		m_owner.m_state_headers.setStatusLine(HttpStatusLine(HttpVersion(http_major, http_minor), code, message));
		m_owner.switchState(HttpStream::STATE_HEADERS);
	}
}

void
HttpStateStartLine::switchToSimpleResponse(string& pool)
{
	pool.assign(m_consumed_data+pool);
	m_owner.m_connection_needs_closing = true;
	HttpResponseMetadata metadata(HttpStatusLine(HttpVersion::HTTP_0_9, 200), HttpHeadersCollection());
	metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
	m_owner.eventGotMetadata(metadata);
	m_owner.switchState(HttpStream::STATE_SIMPLE_RESPONSE);
}

