/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstreamwriter.h"

#include <sstream>
#include "httpheader.h"
#include "httpheaderelement.h"
#include "date.h"
#include "utils.h"

HttpStreamWriter::HttpStreamWriter(Receiver& receiver, const HttpMessageMetadata& metadata,
                                   bool chunked_ok, bool force_conn_close)
:	m_receiver(receiver), m_chunked_output(false)
{
	m_no_body = (metadata.getBodyStatus() == HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
	HttpHeadersCollection headers = metadata.headers();
	if (metadata.getBodyStatus() == HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS) {
		m_conn_close = force_conn_close;
	} else {
		bool body_size_known = (metadata.getBodyStatus() == HttpMessageMetadata::BODY_SIZE_KNOWN);
		if (!body_size_known) {
			headers.removeHeader("Content-Length");
			if (chunked_ok) {
				m_chunked_output = true;
				if (!headers.getHeader("Transfer-Encoding").elements().empty()) {
					headers.setHeader(HttpHeader("Transfer-Encoding", "chunked"));
				} else {
					if (headers.hasHeader("Transfer-Encoding")) {
						headers.getHeaderPtr("Transfer-Encoding")->elements().appendElement(HttpHeaderElement("chunked"));
					} else {
						headers.setHeader(HttpHeader("Transfer-Encoding", "chunked"));
					}
				}
			}
		} else {
			headers.setHeader(HttpHeader("Content-Length", num2str(metadata.getBodySize())));
		}
		m_conn_close = (!m_chunked_output && !body_size_known) || force_conn_close;
		if (m_conn_close) {
			if (headers.hasHeader("Connection")) {
				if (!headers.getHeaderPtr("Connection")->elements().hasElement("close")) {
					headers.getHeaderPtr("Connection")->elements().appendElement(HttpHeaderElement("close"));
				}
			} else {
				headers.setHeader(HttpHeader("Connection", "close"));
			}
		} else {
			if (headers.hasHeader("Connection")) {
				headers.getHeaderPtr("Connection")->elements().removeElementsByName("close");
				if (headers.getHeaderPtr("Connection")->elements().empty()) {
					headers.removeHeader("Connection");
				}
			}
		}
	}
	if (!headers.hasHeader("Date")) {
		headers.setHeader(HttpHeader("Date", Date::formatCurrentTime()));
	}
	m_receiver.moreData(metadata.getStartLine().toString()+"\r\n"+headers.toString()+"\r\n");
}

void
HttpStreamWriter::write(const string& data)
{
	if (data.empty() || m_no_body)
		return;
	if (m_chunked_output) {
		ostringstream strm;
		strm << hex << data.length() << "\r\n";
		strm << data << "\r\n";
		m_receiver.moreData(strm.str());
	} else {
		m_receiver.moreData(data);
	}
}

void
HttpStreamWriter::end()
{
	if (m_chunked_output) {
		m_receiver.moreData("0\r\n\r\n");
	}
	m_receiver.done(m_conn_close);
}

