/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <map>
#include <algorithm>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif
#include "conf.h"
#include "rules.h"
#include "network.h"
#include "errorreporter.h"
#include "childfinishexception.h"
#include "utils.h"
#include "ipcportal.h"
#include "syscall.h"
#include "debug.h"
#include "bfilter.h"

using namespace std;

static void sig_child(int signum)
{
	if (signum != SIGCHLD) {
		return;
	}
	int saved_errno = errno;
	while (waitpid(-1, NULL, WNOHANG) > 0);
	errno = saved_errno;
}

static void sig_term(int signum)
{
	kill(0, SIGTERM);
	while (waitpid(-1, NULL, WNOHANG) > 0);
	exit(0);
}

static void setup_parent_signals()
{
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);
	
	sa.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sa, NULL);
	
	sa.sa_handler = sig_child;
	sa.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &sa, NULL);
}

static void setup_child_signals()
{
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	sa.sa_handler = SIG_DFL;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);
}

#ifndef HAVE_DAEMON
static int daemon(int nocd, int noclose)
{
	struct rlimit rlim;
	pid_t pid;
	int i;
	
	umask(0);
	
	if (getrlimit(RLIMIT_NOFILE, &rlim) < 0)
		perror("getrlimit() failed");
	
	pid = fork();
	if (pid < 0)
		return -1;
	if (pid != 0)
		exit(0);
	
	setsid();
	
	signal(SIGHUP, SIG_IGN);
	pid = fork();
	if (pid < 0)
		return -1;
	if (pid != 0)
		exit(0);
	
	if (!nocd)
		chdir("/");
	
	if (!noclose) {
#if 0
		if (rlim.rlim_max == RLIM_INFINITY)
			rlim.rlim_max = 1024;
		for (i = 0; i < rlim.rlim_max; i++) {
			close(i);
		}
#endif
		i = open("/dev/null", O_RDWR);
		if (i < 0)
			return -1;
		dup2(i, 0);
		dup2(i, 1);
		dup2(i, 2);
		if (i > 2)
			close(i);
	}
	return 0;
}
#endif

int main(int argc, char **argv)
{
	Network::Socket serv_sock, clnt_sock, dbg_serv_sock, dbg_sock = Network::INVALID_SOCK;
	struct sockaddr_in client;
	
#ifdef DEBUG
	ErrorReporter::info(
		"Debugging is enabled. The port 8081 on localhost will be opened for a\n"
		"debug collector to connect to. The debug collector is a GUI program that\n"
		"displays debug info and traffic between BFilter and both the client and server\n"
		"for each connection. I never released it to public, but if you need it,\n"
		"feel free to email me to <joseph_a@mail.ru> and ask for it.\n"
		"Be WARNED that anyone on localhost will be able to connect to port 8081\n"
		"and read all the traffic coming through BFilter."
	);
#endif

	Conf::setConfDir(CONFDIR);
	Rules::setConfDir(CONFDIR);
	Conf::getInstance().readConfig();
	Rules::getInstance().readRules();
	
	NetworkError err;
	serv_sock = Network::tcpServer(Conf::getInstance().getListenAddress().c_str(),
		Conf::getInstance().getListenPort(), &err);
	if (!err.isOK()) {
		ErrorReporter::fatalError(err.getErrorMessage());
	}
#ifdef DEBUG
	dbg_serv_sock = Network::tcpServer("127.0.0.1", 8081, &err);
	if (!err.isOK()) {
		ErrorReporter::fatalError(err.getErrorMessage());
	}
#endif
	
	daemon(1, 0);

	setup_parent_signals();
	
#ifdef DEBUG
	IPCPortal dbg_portal;
	
	map<pid_t, IPCPortal> pid2portal;
	map<int, pid_t> rfd2pid;
	map<int, pid_t> wfd2pid;
	
	fd_set rfds;
	fd_set wfds;
#endif
	while (true) {
#ifdef DEBUG
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(serv_sock, &rfds);
		FD_SET(dbg_serv_sock, &rfds);
		int max = std::max(serv_sock, dbg_serv_sock);
		
		bool dbg_need_write = false;
		if (Network::isValidSocket(dbg_sock)) {
			FD_SET(dbg_sock, &rfds);
			if (dbg_portal.hasDataToSend()) {
				FD_SET(dbg_sock, &wfds);
				dbg_need_write = true;
			}
			if (dbg_sock > max)
				max = dbg_sock;
		}
		
		map<int, pid_t>::const_iterator iter = rfd2pid.begin();
		for (; iter != rfd2pid.end(); ++iter) {
			if (iter->first > max)
				max = iter->first;
			FD_SET(iter->first, &rfds);
		}
		
		iter = wfd2pid.begin();
		for (; iter != wfd2pid.end(); ++iter) {
			if (!pid2portal.find(iter->second)->second.hasDataToSend())
				continue;
			if (iter->first > max)
				max = iter->first;
			FD_SET(iter->first, &wfds);
		}
		
		int ndesc = 0;
		SYSCALL(ndesc = select(max+1, &rfds, &wfds, NULL, NULL));
		
		if (dbg_need_write && FD_ISSET(dbg_sock, &wfds)) {
			if (!dbg_portal.writeChunk()) {
				Network::closeSocket(dbg_sock);
				dbg_sock = Network::INVALID_SOCK;
			}
		}
		
		if (Network::isValidSocket(dbg_sock) && FD_ISSET(dbg_sock, &rfds)) {
			if (!dbg_portal.readChunk()) {
				Network::closeSocket(dbg_sock);
				dbg_sock = Network::INVALID_SOCK;
			}
			while (dbg_portal.nReceived()) {
				IPCMessage msg = dbg_portal.readMessage();
				// do nothing with the message
			}
		}
		
		list<pid_t> remove_list;
		iter = rfd2pid.begin();
		for (; iter != rfd2pid.end(); ++iter) {
			IPCPortal *portal = &pid2portal.find(iter->second)->second;
			if (FD_ISSET(iter->first, &rfds)) {
				if (!portal->readChunk()) {
					// the child has exited or some unexpected error
					remove_list.push_back(iter->second);
				} else {
					while (portal->nReceived()) {
						IPCMessage msg = portal->readMessage();
						// forward the message to debugger
						if (Network::isValidSocket(dbg_sock)) {
							dbg_portal.sendMessage(msg);
						}
					}
				}
			}
		}
		
		iter = wfd2pid.begin();
		for (; iter != wfd2pid.end(); ++iter) {
			IPCPortal *portal = &pid2portal.find(iter->second)->second;
			if (!portal->hasDataToSend())
				continue;
			if (FD_ISSET(iter->first, &wfds)) {
				if (!portal->writeChunk()) {
					// the child has exited or some unexpected error
					remove_list.push_back(iter->second);
				}
			}
		}
		
		list<pid_t>::iterator iter1 = remove_list.begin();
		for (; iter1 != remove_list.end(); ++iter1) {
			IPCPortal *portal = &pid2portal.find(*iter1)->second;
			close(portal->getReadFD());
			close(portal->getWriteFD());
			rfd2pid.erase(portal->getReadFD());
			wfd2pid.erase(portal->getWriteFD());
			pid2portal.erase(*iter1);
		}
		
		remove_list.clear();
		if (FD_ISSET(dbg_serv_sock, &rfds)) {
			if (Network::isValidSocket(dbg_sock)) {
				Network::closeSocket(Network::tcpServerWaitConn(dbg_serv_sock, &client));
				dbg_portal = IPCPortal();
			} else {
				dbg_sock = Network::tcpServerWaitConn(dbg_serv_sock, &client);
				dbg_portal = IPCPortal(dbg_sock, dbg_sock);
			}
		}
		
		if (FD_ISSET(serv_sock, &rfds)) {
#endif // DEBUG
			clnt_sock = Network::tcpServerWaitConn(serv_sock, &client);
			if (!Network::isValidSocket(clnt_sock))
				continue;
#ifdef DEBUG
			int pipe_p2c[2];
			int pipe_c2p[2];
			pipe(pipe_p2c);
			pipe(pipe_c2p);
			int prfd = pipe_c2p[0];
			int pwfd = pipe_p2c[1];
			int crfd = pipe_p2c[0];
			int cwfd = pipe_c2p[1];
#endif
			pid_t pid = fork();
			if (pid == -1) {
				return 1;
			} else if (pid == 0) { // child
				setup_child_signals();
				Network::closeSocket(serv_sock);
				Network::sockSetNodelay(clnt_sock, true);
#ifdef DEBUG
				close(prfd);
				close(pwfd);
				Network::closeSocket(dbg_serv_sock);
				if (Network::isValidSocket(dbg_sock)) {
					Network::closeSocket(dbg_sock);
				}
				
				map<pid_t, IPCPortal>::iterator iter = pid2portal.begin();
				for (; iter != pid2portal.end(); ++iter) {
					close(iter->second.getReadFD());
					close(iter->second.getWriteFD());
				}
				pid2portal.clear();
				rfd2pid.clear();
				wfd2pid.clear();
				
				Debug::setPortal(IPCPortal(crfd, cwfd));
				Debug::registerConnection();
#endif
				try {
					BFilter bfilter(clnt_sock);
					bfilter.run();
				} catch (ChildFinishException& e) {}
				DEBUGLOG("Process finishes");
#ifdef DEBUG
				Debug::unregisterConnection();
#endif
				Network::disconnectAndCloseSocket(clnt_sock, 10);
				return 0;
			} else { // parent
				Network::closeSocket(clnt_sock);
#ifdef DEBUG
				close(crfd);
				close(cwfd);
				map<pid_t, IPCPortal>::iterator fit = pid2portal.find(pid);
				if (fit != pid2portal.end()) {
					close(fit->second.getReadFD());
					close(fit->second.getWriteFD());
					pid2portal.erase(fit);
				}
				pid2portal.insert(map<pid_t, IPCPortal>::value_type(pid, IPCPortal(prfd, pwfd)));
				rfd2pid[prfd] = pid;
				wfd2pid[pwfd] = pid;
#endif
			}
#ifdef DEBUG
		}
#endif
	}
	return 0;
}

