/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <map>
#include <list>
#include <conflexer.h>
#include <confparser.h>
#include "errorreporter.h"
#include "utils.h"
#include "rules.h"

using namespace std;

#define MAX_FILTER       1
#define MAX_AD           2
#define MAX_SCRIPTS      7
#define MAX_JSENGINE     1
#define MAX_TARGET_BLANK 1

Rules Rules::m_instance;
string Rules::m_conf_dir;

Rules::Rules()
{
	m_default_rule.m_filter = 1;
	m_default_rule.m_ad = 0;
	m_default_rule.m_scripts = 3;
	m_default_rule.m_jsengine = 1;
	m_default_rule.m_target_blank = 0;
}

void
Rules::readRules()
{
	const string GLOBAL_RULE_FILE = m_conf_dir+"/rules";
	const string LOCAL_RULE_FILE = m_conf_dir+"/rules.local";
	FILE *fp = fopen(GLOBAL_RULE_FILE.c_str(), "r");
	if (!fp) {
		ErrorReporter::fatalError(string("Could not open ")+GLOBAL_RULE_FILE);
	}
	m_local = false;
	ConfLexer::reset();
	ConfParser::reset();
	ConfParser::setFileName(GLOBAL_RULE_FILE);
	ConfParser::attachListener(this);
	ConfParser::parse(fp);
	ConfParser::detachListener(this);
	fclose(fp);
	processDefaultRule(m_default_global_rule, m_default_rule, m_global_rules);
	
	fp = fopen(LOCAL_RULE_FILE.c_str(), "r");
	if (fp) {
		m_local = true;
		ConfLexer::reset();
		ConfParser::reset();
		ConfParser::setFileName(LOCAL_RULE_FILE);
		ConfParser::attachListener(this);
		ConfParser::parse(fp);
		ConfParser::detachListener(this);
		fclose(fp);
	} else {
		ErrorReporter::warning(string("Could not open ")+LOCAL_RULE_FILE);
	}
	processDefaultRule(m_default_local_rule, m_default_global_rule, m_local_rules);
	
	groupRules();
	
	m_global_rules.clear();
	m_local_rules.clear();
}

int
Rules::getFilterParam(const string& url) const
{
	grouped_regex_type::const_iterator iter = m_local_re_by_filter.begin();
	for (; iter != m_local_re_by_filter.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_filter.begin();
	for (; iter != m_global_re_by_filter.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_filter;
}

int
Rules::getAdParam(const string& url) const
{
	grouped_regex_type::const_iterator iter = m_local_re_by_ad.begin();
	for (; iter != m_local_re_by_ad.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_ad.begin();
	for (; iter != m_global_re_by_ad.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_ad;
}

int
Rules::getScriptsParam(const string& url) const
{
	grouped_regex_type::const_iterator iter = m_local_re_by_scripts.begin();
	for (; iter != m_local_re_by_scripts.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_scripts.begin();
	for (; iter != m_global_re_by_scripts.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_scripts;
}

int
Rules::getJSengineParam(const string& url) const
{
	grouped_regex_type::const_iterator iter = m_local_re_by_jsengine.begin();
	for (; iter != m_local_re_by_jsengine.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_jsengine.begin();
	for (; iter != m_global_re_by_jsengine.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_jsengine;
}

int
Rules::getTargetBlankParam(const string& url) const
{
	grouped_regex_type::const_iterator iter = m_local_re_by_target_blank.begin();
	for (; iter != m_local_re_by_target_blank.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_target_blank.begin();
	for (; iter != m_global_re_by_target_blank.end(); ++iter) {
		if (iter->second.matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_target_blank;
}

void
Rules::processSection(const string& section, int line)
{
	// try compiling the regex just to check the syntax
	regex_t regex;
	int res = regcomp(&regex, string(string("^http://")+section+"$").c_str(), REG_EXTENDED|REG_ICASE|REG_NOSUB);
	if (res) {
		processError("error compiling regex", line);
	} else {
		regfree(&regex);
	}
}

void
Rules::processParameter(const string& section, const string& key, const string& value, int line)
{
	int val = validateParameter(key, value, line);
	if (val < 0)
		return;
	rule_storage_type& storage = (m_local ? m_local_rules : m_global_rules);
	storage[section][key] = val;
}

void
Rules::processError(const string& errmsg, int line)
{
	ErrorReporter::fatalError(string("File ")+ConfParser::getFileName()+"[line "+num2str(line)+"]: "+errmsg);
}

void
Rules::warning(const string& msg, int line)
{
	ErrorReporter::warning(string("File ")+ConfParser::getFileName()+" [line "+num2str(line)+"]: "+msg);
}

int
Rules::validateParameter(const string& key, const string& value, int line)
{
	int range_min = 0;
	int range_max;
	if (key == "filter") {
		range_max = MAX_FILTER;
	} else if (key == "ad") {
		range_max = MAX_AD;
	} else if (key == "scripts") {
		range_max = MAX_SCRIPTS;
	} else if (key == "jsengine") {
		range_max = MAX_JSENGINE;
	} else if (key == "target_blank") {
		range_max = MAX_TARGET_BLANK;
	} else {
		range_max = -1; // prevent compiler warning
		warning(string("unknown key: ")+key, line);
		return -1;
	}
	int val = -1;
	istringstream strm(value);
	strm >> val;
	ostringstream ost;
	ost << val;
	if (ost.str() != value) {
		warning("value must be integer", line);
		return -1;
	}
	if (val < range_min || val > range_max) {
		warning("value outside of range", line);
		return -1;
	}
	return val;
}


void
Rules::processDefaultRule(Rule& result, const Rule& base, rule_storage_type& rules)
{
	rule_storage_type::iterator fit = rules.find("");
	if (fit == rules.end()) {
		result = base;
		return;
	}
	map<string, int>::const_iterator fit1 = fit->second.find("filter");
	if (fit1 == fit->second.end()) {
		result.m_filter = base.m_filter;
	} else {
		result.m_filter = fit1->second;
	}
	fit1 = fit->second.find("ad");
	if (fit1 == fit->second.end()) {
		result.m_ad = base.m_ad;
	} else {
		result.m_ad = fit1->second;
	}
	fit1 = fit->second.find("scripts");
	if (fit1 == fit->second.end()) {
		result.m_scripts = base.m_scripts;
	} else {
		result.m_scripts = fit1->second;
	}
	fit1 = fit->second.find("jsengine");
	if (fit1 == fit->second.end()) {
		result.m_jsengine = base.m_jsengine;
	} else {
		result.m_jsengine = fit1->second;
	}
	fit1 = fit->second.find("target_blank");
	if (fit1 == fit->second.end()) {
		result.m_target_blank = base.m_target_blank;
	} else {
		result.m_target_blank = fit1->second;
	}
	rules.erase(fit);
}

void
Rules::groupRules()
{
	typedef map<string, map<int, list<string> > > storage_type; // [key => [value => [regexes]]]
	storage_type global_storage, local_storage;
	bool local = false;
	do {
		storage_type& storage = (local ? local_storage : global_storage);
		rule_storage_type& source = (local ? m_local_rules : m_global_rules);
		const Rule& def = (local ? m_default_local_rule : m_default_global_rule);
		map<string, int> def_map;
		def_map["filter"] = def.m_filter;
		def_map["ad"] = def.m_ad;
		def_map["scripts"] = def.m_scripts;
		def_map["jsengine"] = def.m_jsengine;
		def_map["target_blank"] = def.m_target_blank;
		rule_storage_type::const_iterator iter1 = source.begin();
		for (; iter1 != source.end(); ++iter1) {
			map<string, int>::const_iterator iter2 = iter1->second.begin();
			for (; iter2 != iter1->second.end(); ++iter2) {
				if (def_map[iter2->first] == iter2->second) {
					/*
					this is an optimisation:
					if the group-by value is equal to default value,
					then there is no need to associate a regex to it
					because it will fall back to default without it
					*/
					continue;
				}
				storage[iter2->first][iter2->second].push_back(iter1->first);
			}
		}
		local = !local;
	} while (local == true);
	
	groupBy("filter", global_storage["filter"], m_global_re_by_filter);
	groupBy("filter", local_storage["filter"], m_local_re_by_filter);
	groupBy("ad", global_storage["ad"], m_global_re_by_ad);
	groupBy("ad", local_storage["ad"], m_local_re_by_ad);
	groupBy("scripts", global_storage["scripts"], m_global_re_by_scripts);
	groupBy("scripts", local_storage["scripts"], m_local_re_by_scripts);
	groupBy("jsengine", global_storage["jsengine"], m_global_re_by_jsengine);
	groupBy("jsengine", local_storage["jsengine"], m_local_re_by_jsengine);
	groupBy("target_blank", global_storage["target_blank"], m_global_re_by_target_blank);
	groupBy("target_blank", local_storage["target_blank"], m_local_re_by_target_blank);
}

void
Rules::groupBy(const string& key, const map<int, list<string> >& src, grouped_regex_type& dst)
{
	map<int, list<string> >::const_iterator iter1 = src.begin();
	for (; iter1 != src.end(); ++iter1) {
		int value = iter1->first;
		if (iter1->second.empty()) {
			dst[value] = Regex();
			continue;
		}
		string composite_regex;
		list<string>::const_iterator iter2 = iter1->second.begin();
		for (; iter2 != iter1->second.end(); ++iter2) {
			if (iter2 != iter1->second.begin()) {
				composite_regex += "|";
			}
			composite_regex += "(^http://"+*iter2+"$)";
		}
		dst[value] = Regex(composite_regex, REG_EXTENDED|REG_ICASE|REG_NOSUB);
		if (dst[value].isNull()) {
			ErrorReporter::fatalError("Could not compile composite regex");
		}
	}
}

