/*
    BFilter - a smart ad-filtering http proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <list>
#include <map>
#include <string>
#include <cstring>
#include <sstream>
#include <algorithm>
#include <cctype>
#include "utils.h"

using namespace std;

void chomp(string& str)
{
	string::size_type len = str.length();
	if (len == 0) return;
	string::size_type pos;
	if ((pos = str.rfind("\n")) == len-1) {
		str.replace(pos, 1, "");
		len--;
	}
	if (len == 0) return;
	if ((pos = str.rfind("\r")) == len-1) {
		str.replace(pos, 1, "");
	}
}

string ltrim(const string& str)
{
	string res;
	int len = str.length();
	for (int i = 0; i < len; i++) {
		if (isspace(str[i])) continue;
		res = str.substr(i);
		break;
	}
	return res;
}

string rtrim(const string& str)
{
	string res;
	int len = str.length();
	for (int i = len-1; i >= 0; i--) {
		if (isspace(str[i])) continue;
		res = str.substr(0, i+1);
		break;
	}
	return res;
}

string trim(const string& str)
{
	return rtrim(ltrim(str));
}

bool string_starts_with(const string& str, const string& prefix)
{
	if (str.length() < prefix.length())
		return false;
	return !memcmp(str.c_str(), prefix.c_str(), prefix.length());
}

bool string_ends_with(const string& str, const string& trailer)
{
	if (str.length() < trailer.length())
		return false;
	return !memcmp(str.c_str()+str.length()-trailer.length(), trailer.c_str(), trailer.length());
}

string join(const string& glue, const list<string>& arr)
{
	string res;
	list<string>::const_iterator iter = arr.begin();
	bool first = true;
	for (; iter != arr.end(); iter++) {
		if (!first) {
			res.append(glue);
		} else first = false;
		res.append(*iter);
	}
	return res;
}

string strtolower(const string& str)
{
	string res(str);
	transform(res.begin(), res.end(), res.begin(), static_cast<int(*)(int)>(tolower));
	return res;
}

string strtoupper(const string& str)
{
	string res(str);
	transform(res.begin(), res.end(), res.begin(), static_cast<int(*)(int)>(toupper));
	return res;
}

string str_replace(const string& search, const string& replace, const string& subject)
{
	string res;
	string::size_type spos = 0;
	while (true) {
		string::size_type pos = subject.find(search, spos);
		if (pos != string::npos) {
			res.append(subject, spos, pos-spos);
			res += replace;
			spos = pos+search.length();
		} else {
			res.append(subject, spos, subject.length()-spos);
			break;
		}
	}
	return res;
}

string htmlencode(const string& str)
{
	string res;
	string::const_iterator iter = str.begin();
	for (; iter != str.end(); iter++) {
		const char ch = *iter;
		if (ch == '<') {
			res += "&lt;";
		} else if (ch == '>') {
			res += "&gt;";
		} else if (ch == '&') {
			res += "&amp;";
		} else if (ch == '"') {
			res += "&quot;";
		} else {
			res += ch;
		}
	}
	return res;
}

static map<string, unsigned char> make_entities()
{
	map<string, unsigned char> ent;
	ent["&amp;"]    = '&';
	ent["&quot;"]   = '"';
	ent["&lt;"]     = '<';
	ent["&gt;"]     = '>';
	ent["&nbsp;"]   = 160;
	ent["&iexcl;"]  = 161;
	ent["&cent;"]   = 162;
	ent["&pound;"]  = 163;
	ent["&curren;"] = 164;
	ent["&yen;"]    = 165;
	ent["&brvbar;"] = 166;
	ent["&sect;"]   = 167;
	ent["&uml;"]    = 168;
	ent["&copy;"]   = 169;
	ent["&ordf;"]   = 170;
	ent["&laquo;"]  = 171;
	ent["&not;"]    = 172;
	ent["&shy;"]    = 173;
	ent["&reg;"]    = 174;
	ent["&macr;"]   = 175;
	ent["&deg;"]    = 176;
	ent["&plusmn;"] = 177;
	ent["&sup2;"]   = 178;
	ent["&sup3;"]   = 179;
	ent["&acute;"]  = 180;
	ent["&micro;"]  = 181;
	ent["&para;"]   = 182;
	ent["&middot;"] = 183;
	ent["&cedil;"]  = 184;
	ent["&sup1;"]   = 185;
	ent["&ordm;"]   = 186;
	ent["&raquo;"]  = 187;
	ent["&frac14;"] = 188;
	ent["&frac12;"] = 189;
	ent["&frac34;"] = 190;
	ent["&iquest;"] = 191;
	ent["&Agrave;"] = 192;
	ent["&Aacute;"] = 193;
	ent["&Acirc;"]  = 194;
	ent["&Atilde;"] = 195;
	ent["&Auml;"]   = 196;
	ent["&Aring;"]  = 197;
	ent["&AElig;"]  = 198;
	ent["&Ccedil;"] = 199;
	ent["&Egrave;"] = 200;
	ent["&Eacute;"] = 201;
	ent["&Ecirc;"]  = 202;
	ent["&Euml;"]   = 203;
	ent["&Igrave;"] = 204;
	ent["&Iacute;"] = 205;
	ent["&Icirc;"]  = 206;
	ent["Iuml;"]    = 207;
	ent["ETH;"]     = 208;
	ent["&Ntilde;"] = 209;
	ent["&Ograve;"] = 210;
	ent["&Oacute;"] = 211;
	ent["&Ocirc;"]  = 212;
	ent["&Otilde;"] = 213;
	ent["&Ouml;"]   = 214;
	ent["&times;"]  = 215;
	ent["&Oslash;"] = 216;
	ent["&Ugrave;"] = 217;
	ent["&Uacute;"] = 218;
	ent["&Ucirc;"]  = 219;
	ent["&Uuml;"]   = 220;
	ent["&Yacute;"] = 221;
	ent["&THORN;"]  = 222;
	ent["&szlig;"]  = 223;
	ent["&agrave;"] = 224;
	ent["&aacute;"] = 225;
	ent["&acirc;"]  = 226;
	ent["&atilde;"] = 227;
	ent["&auml;"]   = 228;
	ent["&aring;"]  = 229;
	ent["&aelig;"]  = 230;
	ent["&ccedil;"] = 231;
	ent["&egrave;"] = 232;
	ent["&eacute;"] = 233;
	ent["&ecirc;"]  = 234;
	ent["&euml;"]   = 235;
	ent["&igrave;"] = 236;
	ent["&iacute;"] = 237;
	ent["&icirc;"]  = 238;
	ent["&iuml;"]   = 239;
	ent["&eth;"]    = 240;
	ent["&ntilde;"] = 241;
	ent["&ograve;"] = 242;
	ent["&oacute;"] = 243;
	ent["&ocirc;"]  = 244;
	ent["&otilde;"] = 245;
	ent["&ouml;"]   = 246;
	ent["&divide;"] = 247;
	ent["&oslash;"] = 248;
	ent["&ugrave;"] = 249;
	ent["&uacute;"] = 250;
	ent["&ucirc;"]  = 251;
	ent["&uuml;"]   = 252;
	ent["&yacute;"] = 253;
	ent["&thorn;"]  = 254;
	return ent;
}
static map<string, unsigned char> entities = make_entities();

string htmldecode(const string& str)
{
	string::size_type pos = 0, pos1 = 0, pos2 = 0, len = str.length();
	string res;
	while (pos < len) {
		pos1 = str.find('&', pos);
		if (pos1 == string::npos) {
			res.append(str, pos, string::npos);
			break;
		} else {
			if (pos1 > pos) {
				res.append(str, pos, pos1-pos);
			}
			pos2 = str.find_first_of("&;", pos1+1);
			if (pos2 == string::npos) {
				res.append(str, pos1, string::npos);
				break;
			} else if (str[pos2] == '&') {
				res.append(str, pos1, pos2-pos1);
				pos = pos2;
			} else {
				// str[pos2] == ';'
				if (str[pos1+1] == '#') {
					unsigned int ch = 256;
					istringstream strm(str.substr(pos1+2, pos2-pos1-2));
					strm >> ch;
					if (ch < 256) {
						res.append(1, ch);
					}
				} else {
					map<string, unsigned char>::const_iterator fit = entities.find(str.substr(pos1, pos2-pos1+1));
					if (fit != entities.end()) {
						res += fit->second;
					}
				}
				pos = pos2+1;
			}
		}
	}
	return res;
}

